from typing import Tuple, List
from enum import Enum

import depthai as dai
import numpy as np
from geometry_msgs.msg import Vector3, Quaternion, Pose2D, Point, Transform
# from std_msgs.msg import Header, ColorRGBA, String
# from visualization_msgs.msg import ImageMarker
from geometry_msgs.msg import TransformStamped
from genpy.rostime import Time
from sensor_msgs.msg import CompressedImage, Image, PointCloud2, PointField, Imu  # s, PointCloud
import std_msgs
from tf.msg import tfMessage

"""
--extra-index-url https://rospypi.github.io/simple/
sensor_msgs
geometry_msgs
std_msgs
genpy
tf2-msgs
tf2-ros
"""

class ImuSyncMethod(Enum):
    LINEAR_INTERPOLATE_ACCEL = 'LINEAR_INTERPOLATE_ACCEL'
    LINEAR_INTERPOLATE_GYRO = 'LINEAR_INTERPOLATE_GYRO'
    COPY = 'COPY'


class DepthAi2Ros1:
    xyz = dict()

    def __init__(self, device: dai.Device) -> None:
        self.start_time = dai.Clock.now()
        self.device = device
        self.imu_packets = []


    def header(self, msg: dai.Buffer) -> std_msgs.msg.Header:
        header = std_msgs.msg.Header()
        ts = (msg.getTimestamp() - self.start_time).total_seconds()
        # secs / nanosecs
        header.stamp = Time(int(ts), (ts % 1) * 1e6)
        try:
            header.frame_id = str(msg.getSequenceNum()) # ImgFrame
        except:
            header.frame_id = str(msg.sequence) # IMUReport
        return header
    def CompressedImage(self, imgFrame: dai.ImgFrame) -> CompressedImage:
        msg = CompressedImage()
        msg.header = self.header(imgFrame)
        msg.format = "jpeg"
        msg.data = np.array(imgFrame.getData()).tobytes()
        return msg

    def Imu(self, imu_packet: dai.IMUPacket, sync_mode: ImuSyncMethod = ImuSyncMethod.LINEAR_INTERPOLATE_ACCEL, linear_accel_cov: float = 0., angular_velocity_cov: float = 0) -> Imu:
        if len(self.imu_packets) > 20:
            self.imu_packets.pop(0)

        self.imu_packets.append(imu_packet)

        if sync_mode != ImuSyncMethod.COPY:
            interp_imu_packets = self.fillImuData_LinearInterpolation(self, sync_mode)
            if len(interp_imu_packets) > 0:
                imu_packet = interp_imu_packets[-1]

        msg = Imu()

        if imu_packet.acceleroMeter is not None:
            msg.header = self.header(imu_packet.acceleroMeter)
            msg.linear_acceleration.x = imu_packet.acceleroMeter.x
            msg.linear_acceleration.y = imu_packet.acceleroMeter.y
            msg.linear_acceleration.z = imu_packet.acceleroMeter.z

        if imu_packet.gyroscope is not None:
            msg.header = self.header(imu_packet.gyroscope)
            msg.angular_velocity.x = imu_packet.gyroscope.x
            msg.angular_velocity.y = imu_packet.gyroscope.y
            msg.angular_velocity.z = imu_packet.gyroscope.z

        msg.orientation.x = 0.0
        msg.orientation.y = 0.0
        msg.orientation.z = 0.0
        msg.orientation.w = 1.0

        msg.orientation_covariance = [-1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
        msg.linear_acceleration_covariance = [linear_accel_cov, 0.0, 0.0, 0.0, linear_accel_cov, 0.0, 0.0, 0.0, linear_accel_cov]
        msg.angular_velocity_covariance = [angular_velocity_cov, 0.0, 0.0, 0.0, angular_velocity_cov, 0.0, 0.0, 0.0, angular_velocity_cov]

        return msg

    def _lerp(self, a, b, t):
        return a * (1.0 - t) + b * t

    def _lerpImu(self, a, b, t):
        res = a.__class__()
        res.x = self._lerp(a.x, b.x, t)
        res.y = self._lerp(a.y, b.y, t)
        res.z = self._lerp(a.z, b.z, t)
        return res
    
    def fillImuData_LinearInterpolation(self, sync_mode: ImuSyncMethod):
        accel_hist = []
        gyro_hist = []
        interp_imu_packets = []

        for i in range(len(self.imu_packets)):
            if len(accel_hist) == 0:
                accel_hist.append(self.imu_packets[i].acceleroMeter)
            elif accel_hist[-1].sequence != self.imu_packets[i].acceleroMeter.sequence:
                accel_hist.append(self.imu_packets[i].acceleroMeter)

            if len(gyro_hist) == 0:
                gyro_hist.append(self.imu_packets[i].gyroscope)
            elif gyro_hist[-1].sequence != self.imu_packets[i].gyroscope.sequence:
                gyro_hist.append(self.imu_packets[i].gyroscope)

            if sync_mode.value == ImuSyncMethod.LINEAR_INTERPOLATE_ACCEL:
                if len(accel_hist) < 3:
                    continue
                else:
                    accel0 = dai.IMUReportAccelerometer()
                    accel0.sequence = -1

                    while len(accel_hist) > 0:
                        if accel0.sequence == -1:
                            accel0 = accel_hist.pop(0)
                        else:
                            accel1 = accel_hist.pop(0)
                            dt = (accel1.timestamp.get() - accel0.timestamp.get()).total_seconds() * 1000

                            while len(gyro_hist) > 0:
                                curr_gyro = gyro_hist[0]

                                if curr_gyro.timestamp.get() > accel0.timestamp.get() and curr_gyro.timestamp.get() <= accel1.timestamp.get():
                                    diff = (curr_gyro.timestamp.get() - accel0.timestamp.get()).total_seconds() * 1000
                                    alpha = diff / dt
                                    interp_accel = self._lerpImu(accel0, accel1, alpha)
                                    imu_packet = dai.IMUPacket()
                                    imu_packet.acceleroMeter = interp_accel
                                    imu_packet.gyroscope = curr_gyro
                                    interp_imu_packets.append(imu_packet)
                                    gyro_hist.pop(0)

                                elif curr_gyro.timestamp.get() > accel1.timestamp.get():
                                    accel0 = accel1
                                    if len(accel_hist) > 0:
                                        accel1 = accel_hist.pop(0)
                                        dt = (accel1.timestamp.get() - accel0.timestamp.get()).total_seconds() * 1000
                                    else:
                                        break
                                else:
                                    gyro_hist.pop(0)

                            accel0 = accel1

                    accel_hist.append(accel0)

            elif sync_mode == ImuSyncMethod.LINEAR_INTERPOLATE_GYRO:
                if len(gyro_hist) < 3:
                    continue
                else:
                    gyro0 = dai.IMUReportGyroscope()
                    gyro0.sequence = -1

                    while len(gyro_hist) > 0:
                        if gyro0.sequence == -1:
                            gyro0 = gyro_hist.pop(0)
                        else:
                            gyro1 = gyro_hist.pop(0)
                            dt = (gyro1.timestamp.get() - gyro0.timestamp.get()).total_seconds() * 1000

                            while len(accel_hist) > 0:
                                curr_accel = accel_hist[0]

                                if curr_accel.timestamp.get() > gyro0.timestamp.get() and curr_accel.timestamp.get() <= gyro1.timestamp.get():
                                    diff = (curr_accel.timestamp.get() - gyro0.timestamp.get()).total_seconds() * 1000
                                    alpha = diff / dt
                                    interp_gyro = self._lerpImu(gyro0, gyro1, alpha)
                                    imu_packet = dai.IMUPacket()
                                    imu_packet.acceleroMeter = curr_accel
                                    imu_packet.gyroscope = interp_gyro
                                    interp_imu_packets.append(imu_packet)
                                    accel_hist.pop(0)

                                elif curr_accel.timestamp.get() > gyro1.timestamp.get():
                                    gyro0 = gyro1

                                    if len(gyro_hist) > 0:
                                        gyro1 = gyro_hist.pop(0)
                                        dt = (gyro1.timestamp.get() - gyro0.timestamp.get()).total_seconds() * 1000
                                    else:
                                        break

                                else:
                                    accel_hist.pop(0)

                            gyro0 = gyro1

                    gyro_hist.append(gyro0)

        return interp_imu_packets


    def Image(self, imgFrame: dai.ImgFrame) -> Image:
        msg = Image()
        msg.header = self.header(imgFrame)
        msg.height = imgFrame.getHeight()
        msg.width = imgFrame.getWidth()
        msg.step = imgFrame.getWidth()
        msg.is_bigendian = 0

        type = imgFrame.getType()
        TYPE = dai.ImgFrame.Type
        if type == TYPE.RAW16: # Depth
            msg.encoding = 'mono16'
            msg.step *= 2 # 2 bytes per pixel
            msg.data = imgFrame.getData().tobytes()
        elif type in [TYPE.GRAY8, TYPE.RAW8]: # Mono frame
            msg.encoding = 'mono8'
            msg.data = imgFrame.getData().tobytes() #np.array(imgFrame.getFrame()).tobytes()
        else:
            msg.encoding = 'bgr8'
            msg.data = np.array(imgFrame.getCvFrame()).tobytes()
        return msg

    def TfMessage(self,
           imgFrame: dai.ImgFrame,
           translation: Tuple[float,float,float] = (0., 0., 0.),
           rotation: Tuple[float,float,float, float] = (0., 0., 0., 0.)) -> tfMessage:
        msg = tfMessage()
        tf = TransformStamped()
        tf.header = self.header(imgFrame)
        tf.child_frame_id = str(imgFrame.getSequenceNum())
        tf.transform = Transform(
            translation = Vector3(x=translation[0], y=translation[1], z=translation[2]),
            rotation = Quaternion(x=rotation[0], y=rotation[1], z=rotation[2], w=rotation[3])
        )
        msg.transforms.append(tf)
        return msg

    def PointCloud2(self, imgFrame: dai.ImgFrame):
        """
        Depth frame -> ROS1 PointCloud2 message
        """
        msg = PointCloud2()
        msg.header = self.header(imgFrame)

        heigth = str(imgFrame.getHeight())
        if heigth not in self.xyz:
            self._create_xyz(imgFrame.getWidth(), imgFrame.getHeight())

        frame = imgFrame.getCvFrame()
        frame = np.expand_dims(np.array(frame), axis=-1)
        pcl = self.xyz[heigth] * frame / 1000.0  # To meters

        msg.height = imgFrame.getHeight()
        msg.width = imgFrame.getWidth()
        msg.fields = [
            PointField(name="x", offset=0, datatype=PointField.FLOAT32, count=1),
            PointField(name="y", offset=4, datatype=PointField.FLOAT32, count=1),
            PointField(name="z", offset=8, datatype=PointField.FLOAT32, count=1)
        ]
        msg.is_bigendian = False
        msg.point_step = 12  # 3 * float32 (=4 bytes)
        msg.row_step = 12 * imgFrame.getWidth()
        msg.data = pcl.tobytes()
        msg.is_dense = True
        return msg

    def _create_xyz(self, width, height):
        calibData = self.device.readCalibration()
        M_right = calibData.getCameraIntrinsics(dai.CameraBoardSocket.RIGHT, dai.Size2f(width, height))
        camera_matrix = np.array(M_right).reshape(3, 3)

        xs = np.linspace(0, width - 1, width, dtype=np.float32)
        ys = np.linspace(0, height - 1, height, dtype=np.float32)

        # generate grid by stacking coordinates
        base_grid = np.stack(np.meshgrid(xs, ys))  # WxHx2
        points_2d = base_grid.transpose(1, 2, 0)  # 1xHxWx2

        # unpack coordinates
        u_coord: np.array = points_2d[..., 0]
        v_coord: np.array = points_2d[..., 1]

        # unpack intrinsics
        fx: np.array = camera_matrix[0, 0]
        fy: np.array = camera_matrix[1, 1]
        cx: np.array = camera_matrix[0, 2]
        cy: np.array = camera_matrix[1, 2]

        # projective
        x_coord: np.array = (u_coord - cx) / fx
        y_coord: np.array = (v_coord - cy) / fy

        xyz = np.stack([x_coord, y_coord], axis=-1)
        self.xyz[str(height)] = np.pad(xyz, ((0, 0), (0, 0), (0, 1)), "constant", constant_values=1.0)



import importlib
from enum import IntEnum
from pathlib import Path
import os
from typing import Dict, Union, Optional, Tuple
import requests
import depthai as dai

BLOBS_PATH = Path.home() / Path('.cache/blobs')


class AspectRatioResizeMode(IntEnum):
    """
    If NN input frame is in different aspect ratio than what the model expect, we have 3 different
    modes of operation. Full documentation:
    https://docs.luxonis.com/projects/api/en/latest/tutorials/maximize_fov/
    """
    LETTERBOX = 0  # Preserves full FOV, but smaller frame means less features which might decrease NN accuracy
    STRETCH = 1  # Preserves full FOV, but frames are stretched which might decrease NN accuracy
    CROP = 2  # Crops some FOV to match the required FOV. No potential NN accuracy decrease.


def getBlob(url: str) -> Path:
    """
    Download the blob path from the url. If blob is cached, serve that. TODO: compute hash, check server hash,
    as there will likely be many `model.blob`s.

    @param url: Url to the blob
    @return: Local path to the blob
    """
    fileName = Path(url).name
    filePath = BLOBS_PATH / fileName
    if filePath.exists():
        return filePath
    BLOBS_PATH.mkdir(parents=True, exist_ok=True)

    r = requests.get(url)
    with open(filePath, 'wb') as f:
        f.write(r.content)
        print('Downloaded', fileName)

    return filePath


# Copied from utils.py - remove that once DepthAI Demo is deprecated
def loadModule(path: Path):
    """
    Loads module from specified path. Used internally e.g. to load a custom handler file from path

    Args:
        path (pathlib.Path): path to the module to be loaded

    Returns:
        module: loaded module from provided path
    """
    spec = importlib.util.spec_from_file_location(path.stem, str(path.absolute()))
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


# Copid from utils.py (due to circular import)
def isUrl(source: Union[str, Path]) -> bool:
    if isinstance(source, Path):
        source = str(source)
    return source.startswith("http://") or source.startswith("https://")


def getSupportedModels(printModels=True) -> Dict[str, Path]:
    folder = Path(os.path.dirname(__file__)).parent / "nn_models"
    dic = dict()
    for item in folder.iterdir():
        if item.is_dir() and item.name != '__pycache__':
            dic[item.name] = item

    if printModels:
        print("\nDepthAI SDK supported models:\n")
        [print(f"- {name}") for name in dic]
        print('')
    return dic



import argparse
from pathlib import Path
from typing import Any, Dict

import depthai as dai

from depthai_sdk.components.parser import rgb_resolution, mono_resolution, parse_bool

"""
Very similar to ArgsManager, but specific to the new SDK, not depthai_demo.py. 
"""

# TODO: support changing colorMaps
# try:
#     import cv2
#     _colorMaps = list(
#         map(lambda name: name[len("COLORMAP_"):], filter(lambda name: name.startswith("COLORMAP_"), vars(cv2))))
# except:
#     _colorMaps = None

_openvinoVersions = [v.replace("VERSION_", "") for v in vars(dai.OpenVINO.Version) if v.startswith("VERSION_")]


def _checkRange(minVal, maxVal):
    def checkFn(value):
        ivalue = int(value)
        if minVal <= ivalue <= maxVal:
            return ivalue
        else:
            raise argparse.ArgumentTypeError(
                "{} is an invalid int value, must be in range {}..{}".format(value, minVal, maxVal)
            )

    return checkFn


def _commaSeparated(default, cast=str):
    def _fun(option):
        optionList = option.split(",")
        if len(optionList) not in [1, 2]:
            raise argparse.ArgumentTypeError(
                "{0} format is invalid. See --help".format(option)
            )
        elif len(optionList) == 1:
            return optionList[0], cast(default)
        else:
            try:
                return optionList[0], cast(optionList[1])
            except ValueError:
                raise argparse.ArgumentTypeError(
                    "In option: {0} {1} is not in a correct format!".format(option, optionList[1])
                )

    return _fun


def _checkEnum(enum):
    def _fun(value: str):
        try:
            return getattr(enum, value.upper())
        except:
            choices = [f"'{str(item).split('.')[-1]}'" for name, item in vars(enum).items() if name.isupper()]
            raise argparse.ArgumentTypeError(
                "{} option wasn't found in {} options! Choices: {}".format(value, enum, ', '.join(choices))
            )

    return _fun


class ArgsParser:
    @staticmethod
    def parseArgs(parser: argparse.ArgumentParser = None) -> Dict[str, Any]:
        """
        Creates Argument parser for common OAK device configuration

        Args:
            parser (argparse.ArgumentParser, optional): Use an existing parser. By default it creates a new ArgumentParser.
        """
        if parser is None:
            parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter)

        parser.add_argument('-recording', '--recording', type=str,
                            help="Path or url to a depthai-recording/folder/YouTube url/file to be used instead of live camera feed")

        # Color/Mono cam related arguments
        parser.add_argument("-rgbr", "--rgbResolution", type=str,
                            help="RGB cam res height: (1920x)1080, (3840x)2160, (4056x)3040, (1280x)720, (1280x)800. Default: %(default)s")
        parser.add_argument("-rgbf", "--rgbFps", type=float,
                            help="RGB cam fps: max 118.0 for H:1080, max 42.0 for H:2160. Default: %(default)s")
        parser.add_argument("-monor", "--monoResolution", type=str,
                            help="Mono cam res height: (1280x)720, (1280x)800 or (640x)400. Default: %(default)s")
        parser.add_argument("-monof", "--monoFps", type=float,
                            help="Mono cam fps: max 60.0 for H:720 or H:800, max 120.0 for H:400. Default: %(default)s")
        parser.add_argument('-fps', '--fps', type=float, help='Camera FPS applied to all sensors')

        # ColorCamera ISP values
        parser.add_argument('-isp', '--ispScale', type=_commaSeparated(None), help="Sets ColorCamera's ISP scale")
        parser.add_argument('-sharpness', '--sharpness', type=_checkRange(0, 4),
                            help="Sets ColorCamera's sharpness")
        parser.add_argument('-lumaDenoise', '--lumaDenoise', type=_checkRange(0, 4),
                            help="Sets ColorCamera's Luma denoise")
        parser.add_argument('-chromaDenoise', '--chromaDenoise', type=_checkRange(0, 4),
                            help="Sets ColorCamera's Chroma denoise")

        # ColorCamera controls
        parser.add_argument('-manualFocus', '--manualFocus', type=_checkRange(0, 255),
                            help="Specify a Lens Position between 0 and 255 to use manual focus. Otherwise, auto-focus is used by default.")
        parser.add_argument('-afMode', '--afMode', type=_checkEnum(dai.CameraControl.AutoFocusMode),
                            help="Specify the Auto Focus mode for the ColorCamera. AUTO by default.")
        parser.add_argument('-awbMode', '--awbMode',
                            type=_checkEnum(dai.CameraControl.AutoWhiteBalanceMode),
                            help="Specify the Auto White Balance mode for the ColorCamera. AUTO by default.")
        parser.add_argument('-sceneMode', '--sceneMode', type=_checkEnum(dai.CameraControl.SceneMode),
                            help="Specify the Scene mode for the ColorCamera. AUTO by default.")
        parser.add_argument('-abMode', '-antiBandingMode', '--antiBandingMode',
                            type=_checkEnum(dai.CameraControl.AntiBandingMode),
                            help="Specify the Anti-Banding mode for the ColorCamera. AUTO by default.")
        parser.add_argument('-effectMode', '--effectMode', type=_checkEnum(dai.CameraControl.EffectMode),
                            help="Specify the Effect mode for the ColorCamera. AUTO by default.")

        # parser.add_argument("--cameraControls",type=parse_bool,
        #                     help="Show camera configuration options in GUI and control them using keyboard")
        # parser.add_argument("--cameraExposure", type=_commaSeparated("all", int), nargs="+",
        #                     help="Specify camera saturation")
        # parser.add_argument("--cameraSensitivity", type=_commaSeparated("all", int), nargs="+",
        #                     help="Specify camera sensitivity")
        # parser.add_argument("--cameraSaturation", type=_commaSeparated("all", int), nargs="+",
        #                     help="Specify image saturation")
        # parser.add_argument("--cameraContrast", type=_commaSeparated("all", int), nargs="+",
        #                     help="Specify image contrast")
        # parser.add_argument("--cameraBrightness", type=_commaSeparated("all", int), nargs="+",
        #                     help="Specify image brightness")
        # parser.add_argument("--cameraSharpness", type=_commaSeparated("all", int), nargs="+",
        #                     help="Specify image sharpness")

        # Depth related arguments
        parser.add_argument("-dct", "--disparityConfidenceThreshold", type=_checkRange(0, 255),
                            help="Disparity confidence threshold, used for depth measurement.")
        parser.add_argument("-lrct", "--lrcThreshold", type=_checkRange(0, 10),
                            help="Left right check threshold, used for depth measurement. Default: %(default)s")
        parser.add_argument("-sig", "--sigma", type=_checkRange(0, 250),
                            help="Sigma value for Bilateral Filter applied on depth. Default: %(default)s")
        parser.add_argument("-med", "--stereoMedianSize", type=int, choices=[0, 3, 5, 7],
                            help="Disparity / depth median filter kernel size (N x N) . 0 = filtering disabled. Default: %(default)s")
        parser.add_argument('-lrc', '--stereoLrCheck', type=parse_bool,
                            help="Set stereo LR-check feature.")
        parser.add_argument('-ext', '--extendedDisparity', type=parse_bool,
                            help="Set stereo Extended Disparity feature.")
        parser.add_argument('-sub', '--subpixel', type=parse_bool,
                            help="Set stereo Subpixel feature.")

        # parser.add_argument("-cm", "--colorMap",ET", choices=_colorMaps,
        #                     help="Change color map used to apply colors to depth/disparity frames. Default: %(default)s")

        # Spatial image detection related arguments
        parser.add_argument("-maxd", "--maxDepth", type=int,
                            help="Maximum depth distance for spatial coordinates in mm. Default: %(default)s")
        parser.add_argument("-mind", "--minDepth", type=int,
                            help="Minimum depth distance for spatial coordinates in mm. Default: %(default)s")
        # parser.add_argument('-sbb', '--spatialBoundingBox', type=parse_bool,
        #                     help="Display spatial bounding box (ROI) when displaying spatial information. The Z coordinate get's calculated from the ROI (average)")
        parser.add_argument("-sbbsf", "--sbbScaleFactor", type=float,
                            help="Spatial bounding box scale factor. Sometimes lower scale factor can give better depth (Z) result. Default: %(default)s")

        # Pipeline
        parser.add_argument("--openvinoVersion", type=str, choices=_openvinoVersions,
                            help="Specify which OpenVINO version to use in the pipeline")
        parser.add_argument('-xls', '--xlinkChunkSize', type=int, help="Specify XLink chunk size")
        parser.add_argument('-tun', '--cameraTuning', type=Path,
                            help="Path to camera tuning blob to use, overriding the built-in tuning")

        # Device
        parser.add_argument("-dev", "--deviceId", type=str,
                            help="DepthAI MX id of the device to connect to. Use the word 'list' to show all devices and exit.")
        parser.add_argument('-usbs', '--usbSpeed', type=str, choices=["usb2", "usb3"],
                            help="Force USB communication speed. Default: %(default)s")
        # Device after booting
        parser.add_argument("--irDotBrightness", "-laser", "--laser", type=_checkRange(0, 1200),
                            help="For OAK-D Pro: specify IR dot projector brightness, range: 0..1200 [mA], default 0 (turned off)")
        parser.add_argument("--irFloodBrightness", "-led", "--led", type=_checkRange(0, 1500),
                            help="For OAK-D Pro: specify IR flood illumination brightness, range: 0..1500 [mA], default 0 (turned off)")

        args = parser.parse_args()
        # Parse arguments
        args.rgbResolution = rgb_resolution(args.rgbResolution)
        args.monoResolution = mono_resolution(args.monoResolution)
        # Global FPS setting, applied to all cameras
        if args.fps is not None:
            args.rgbFps = args.fps
            args.monoFps = args.fps

        args = vars(args)  # Namespace->Dict
        # Print user-defined arguments
        for name, val in args.items():
            if val is not None:
                print(f'{name}: {val}')
        return args



import numpy as np
from depthai import NNData

from depthai_sdk import SemanticSegmentation

NN_WIDTH, NN_HEIGHT = 513, 513


def decode(nn_data: NNData) -> SemanticSegmentation:
    mask = np.array(nn_data.getFirstLayerInt32()).reshape(NN_WIDTH, NN_HEIGHT)
    return SemanticSegmentation(nn_data, mask)



from typing import List

import depthai as dai

from depthai_sdk.components.component import Component, XoutBase
from depthai_sdk.oak_outputs.xout import XoutIMU
from depthai_sdk.oak_outputs.xout_base import StreamXout


class IMUComponent(Component):
    node: dai.node.IMU

    def __init__(self, pipeline: dai.Pipeline):
        self.out = self.Out(self)

        super().__init__()
        self.node = pipeline.createIMU()
        self.config_imu()  # Default settings, component won't work without them

    def config_imu(self,
                   sensors: List[dai.IMUSensor] = None,
                   report_rate: int = 100,
                   batch_report_threshold: int = 1,
                   max_batch_reports: int = 10,
                   enable_firmware_update: bool = False) -> None:
        """
        Configure IMU node.

        Args:
            sensors: List of sensors to enable.
            report_rate: Report rate in Hz.
            batch_report_threshold: Number of reports to batch before sending them to the host.
            max_batch_reports: Maximum number of batched reports to send to the host.
            enable_firmware_update: Enable firmware update if true, disable otherwise.

        Returns: None
        """
        sensors = sensors or [dai.IMUSensor.ACCELEROMETER_RAW, dai.IMUSensor.GYROSCOPE_RAW]

        # TODO: check whether we have IMU (and which sensors are supported) once API supports it

        self.node.enableIMUSensor(sensors=sensors, reportRate=report_rate)
        self.node.setBatchReportThreshold(batchReportThreshold=batch_report_threshold)
        self.node.setMaxBatchReports(maxBatchReports=max_batch_reports)
        self.node.enableFirmwareUpdate(enable_firmware_update)

    def _update_device_info(self, pipeline: dai.Pipeline, device: dai.Device, version: dai.OpenVINO.Version):
        pass

    class Out:
        _comp: 'IMUComponent'

        def __init__(self, imuComponent: 'IMUComponent'):
            self._comp = imuComponent

        def main(self, pipeline: dai.Pipeline, device: dai.Device) -> XoutBase:
            """
            Default output. Uses either camera(), replay(), or encoded() depending on the component settings.
            """
            return self.text(pipeline, device)

        def text(self, pipeline: dai.Pipeline, device: dai.Device) -> XoutBase:
            out = self._comp.node.out
            out = StreamXout(self._comp.node.id, out)
            imu_out = XoutIMU(out)
            return self._comp._create_xout(pipeline, imu_out)

    out: Out



#!/usr/bin/python

# (c) 2017, Johnny Chan
# https://github.com/jchanvfx/NodeGraphQt

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of the Johnny Chan nor the names of its contributors
#   may be used to endorse or promote products derived from this software
#   without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
# OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



from marshmallow import Schema, fields
from depthai_sdk.classes.yolo_config import YoloConfig


class Mappings(Schema):
    # List of object detection labels. Labels can either be string, or array of [label, color] where color is in #RGB
    labels = fields.List(fields.Raw)


class Model(Schema):
    # A string, either path or url to .blob
    blob = fields.Str(required=False)
    # To download the model with blobconverter from either the open_model_zoo or the depthai-model-zoo
    model_name = fields.Str(required=False)
    zoo = fields.Str(required=False)
    # Model in IR format, uses blobconverter to convert the model to .blob.
    xml = fields.Str(required=False)
    bin = fields.Str(required=False)


class NNConfig(Schema):
    NN_family = fields.Str(required=False)
    confidence_threshold = fields.Float(required=False)
    NN_specific_metadata = fields.Nested(YoloConfig, required=False)
    output_format = fields.Str()
    input_size = fields.Str()


class Config(Schema):
    # Specifies where the model is/how to download it. It should contain one of the following
    model = fields.Nested(Model)
    # Path to the python script that contains decode() function that decodes dai.NNData into standardized NN results
    handler = fields.Str(required=False)
    nn_config = fields.Nested(NNConfig)  # Contains NN configuration data
    openvino_version = fields.Str(required=False)  # If the model requires specific OpenVINO version
    mappings = fields.Nested(Mappings, required=False)
    version = fields.Int()  # Version of the config file



import os
import time
from threading import Thread
from time import monotonic

import depthai as dai

from depthai_sdk.readers.abstract_reader import AbstractReader
from depthai_sdk.utils import *

_fileTypes = ['color', 'left', 'right', 'disparity', 'depth']
_videoExt = ['.mjpeg', '.avi', '.mp4', '.h265', '.h264', '.webm']
_imageExt = ['.bmp', '.dib', '.jpeg', '.jpg', '.jpe', '.jp2', '.png', '.webp', '.pbm', '.pgm', '.ppm', '.pxm',
             '.pnm', '.pfm', '.sr', '.ras', '.tiff', '.tif', '.exr', '.hdr', '.pic']


class Replay:
    disabledStreams: List[str] = []
    # Nodes
    left: dai.node.XLinkIn = None
    right: dai.node.XLinkIn = None
    color: dai.node.XLinkIn = None

    _inputQueues = dict()  # dai.InputQueue dictionary for each stream
    _seqNum = 0  # Frame sequence number, added to each imgFrame
    _now: monotonic = None
    _colorSize = None
    _keepAR = True  # By default, crop image as needed to keep the aspect ratio
    _pause = False
    _calibData = None

    fps: float = 30.0
    thread: Thread = None
    _stop: bool = False  # Stop the thread that's sending frames to the OAK camera

    xins: List[str] = []  # Name of XLinkIn streams

    reader: AbstractReader = None
    frames: Dict[str, np.ndarray] = dict()  # Frames read from Readers
    imgFrames: Dict[str, dai.ImgFrame] = dict()  # Last frame sent to the device

    def __init__(self, path: str):
        """
        Helper file to replay recorded depthai stream. It reads from recorded files (mjpeg/avi/mp4/h265/h264/bag)
        and sends frames back to OAK camera to replay the scene, including depth reconstruction from 2 synced mono
        streams.
    
        Args:
            path (str): Path to the recording folder
        """
        self.path = self._get_path(path)

        def cntFilesExt(path: Path, ext: Union[str, List[str]]) -> int:
            def fileWithExt(file: str) -> bool:
                if isinstance(ext, List):
                    return os.path.splitext(file)[1] in ext
                elif isinstance(ext, str):
                    return file.endswith(ext)
                else:
                    raise ValueError('ext should be either str or List[str]!')

            return [fileWithExt(f) for f in os.listdir(str(path))].count(True)

        if self.path.is_dir():  # Provided path is a folder
            if 0 < cntFilesExt(self.path, _imageExt):
                from .readers.image_reader import ImageReader
                self.reader = ImageReader(self.path)
            elif 0 < cntFilesExt(self.path, _videoExt):
                from .readers.videocap_reader import VideoCapReader
                self.reader = VideoCapReader(self.path)
            elif cntFilesExt(self.path, '.bag') == 1:
                from .readers.rosbag_reader import RosbagReader
                self.reader = RosbagReader(self.path)
            elif cntFilesExt(self.path, '.mcap') == 1:
                from .readers.mcap_reader import McapReader
                self.reader = McapReader(self.path)
            elif cntFilesExt(self.path, '.db3') == 1:
                from .readers.db3_reader import Db3Reader
                self.reader = Db3Reader(self.path)
            else:
                raise RuntimeError("Path invalid - no recordings found.")

            # Read calibration file
            calibFile = self.path / 'calib.json'
            if calibFile.exists():
                self._calibData = dai.CalibrationHandler(str(calibFile))

        else:  # Provided path is a file
            if self.path.suffix in _videoExt:
                from .readers.videocap_reader import VideoCapReader
                self.reader = VideoCapReader(self.path)
            elif self.path.suffix in _imageExt:
                from .readers.image_reader import ImageReader
                self.reader = ImageReader(self.path)
            else:
                raise NotImplementedError('Please select folder')

    def _get_path(self, path: str) -> Path:
        """
        Either use local depthai-recording, YT link, mp4 url
        @param path: depthai-recording path.
        @return: Replay module
        """
        if isUrl(path):
            if isYoutubeLink(path):
                # Overwrite source - so Replay class can use it
                return downloadYTVideo(path)
            else:
                return downloadContent(path)

        if Path(path).resolve().exists():
            return Path(path).resolve()

        recordingName: str = path
        # Check if we have it stored locally
        path: Path = getLocalRecording(recordingName)
        if path is not None:
            return path
        # Try to download from the server
        dic = getAvailableRecordings()
        if recordingName in dic:
            arr = dic[recordingName]
            print("Downloading depthai recording '{}' from Luxonis' servers, in total {:.2f} MB".format(recordingName,
                                                                                                        arr[1] / 1e6))
            path = downloadRecording(recordingName, arr[0])
            return path
        else:
            raise ValueError(f"DepthAI recording '{recordingName}' was not found on the server!")

    def togglePause(self):
        """
        Toggle pausing of sending frames to the OAK camera.
        """
        self._pause = not self._pause
        print("PAUSE", self._pause)

    def setFps(self, fps: float):
        """
        Sets frequency at which Replay module will send frames to the camera. Default 30FPS.
        """
        self.fps = fps

    def getFps(self) -> float:
        return self.fps

    def setResizeColor(self, size: tuple):
        """
        Resize color frames prior to sending them to the device.

        Args:
            size (tuple(width, heigth)): Size of color frames that are sent to the camera
        """
        self._colorSize = size

    def keepAspectRatio(self, keepAspectRatio: bool):
        """
        Used when we want to resize color frames before sending them to the host. By default,
        this is set to True, so frames are cropped to keep the original aspect ratio.
        """
        self._keepAR = keepAspectRatio

    def disableStream(self, streamName: str, disableReading: bool = False):
        """
        Disable sending a recorded stream to the device.

        Args:
            streamName(str): Name of the stream to disable (eg. 'left', 'color', 'depth', etc.)
            disableReading (bool, Optional): Also disable reading frames from the file
        """
        # if streamName not in self.readers:
        #     print(f"There's no stream '{streamName}' available!")
        #     return
        if disableReading:
            self.reader.disableStream(streamName)

        self.disabledStreams.append(streamName)

    def sendFrames(self, cb=None) -> bool:
        """
        Reads and sends recorded frames from all enabled streams to the OAK camera.

        Returns:
            bool: True if successful, otherwise False.
        """
        if not self._pause:  # If replaying is paused, don't read new frames
            if not self._readFrames():
                return False  # End of the recording

        self._now = monotonic()
        for name in self.frames:
            imgFrame = self._createImgFrame(name, self.frames[name])
            # Save the imgFrame
            self.imgFrames[name] = imgFrame
            if cb:  # callback
                cb(name, imgFrame)

            # Don't send these frames to the OAK camera
            if name in self.disabledStreams: continue

            # Send an imgFrame to the OAK camera
            self._inputQueues[name].send(imgFrame)

        self._seqNum += 1
        return True

    def initPipeline(self, pipeline: dai.Pipeline = None):
        """
        Prepares the pipeline for replaying. It creates XLinkIn nodes and sets up StereoDepth node.
        Returns: dai.Pipeline
        """
        if pipeline is None:  # Create pipeline if not passed
            pipeline = dai.Pipeline()

        if self._calibData is not None:
            pipeline.setCalibrationData(self._calibData)

        def createXIn(p: dai.Pipeline, name: str):
            xin = p.create(dai.node.XLinkIn)
            xin.setMaxDataSize(self._getMaxSize(name))
            xin.setStreamName(name + '_in')
            self.xins.append(name)
            return xin

        for name in self.reader.getStreams():
            if name not in self.disabledStreams:
                xin = createXIn(pipeline, name)
                if name.upper() == 'LEFT':
                    self.left = xin
                elif name.upper() == 'RIGHT':
                    self.right = xin
                elif name.upper() == 'COLOR':
                    self.color = xin
                else:
                    pass  # Not implemented

        return pipeline

    def initStereoDepth(self, stereo: dai.node.StereoDepth):
        streams = self.reader.getStreams()
        if 'left' not in streams or 'right' not in streams:
            raise Exception("Tried to init StereoDepth, but left/right streams aren't available!")
        stereo.setInputResolution(self.getShape('left'))

        if self.color:  # Enable RGB-depth alignment
            stereo.setDepthAlign(dai.CameraBoardSocket.RGB)
            if self._colorSize is not None:
                stereo.setOutputSize(*self._colorSize)
            else:
                stereo.setOutputSize(*self.getShape('color'))

        self.left.out.link(stereo.left)
        self.right.out.link(stereo.right)

    def start(self, cb):
        """
        Start sending frames to the OAK device on a new thread
        """
        self.thread = Thread(target=self.run, args=(cb,))
        self.thread.start()

    def run(self, cb):
        delay = 1.0 / self.fps
        while True:
            if not self.sendFrames(cb): break
            time.sleep(delay)
            if self._stop: break
        print('Replay `run` thread stopped')
        self._stop = True

    def createQueues(self, device: dai.Device):
        """
        Creates input queue for each enabled stream
        
        Args:
            device (dai.Device): Device to which we will stream frames
        """
        for name in self.xins:
            self._inputQueues[name] = device.getInputQueue(name + '_in')

    def getStreams(self) -> List[str]:
        streams: List[str] = []
        [streams.append(name) for name in self.reader.getStreams()]
        return streams

    def _resizeColor(self, frame):
        if self._colorSize is None:
            # No resizing needed
            return frame

        if not self._keepAR:
            # No need to keep aspect ratio, image will be squished
            return cv2.resize(frame, self._colorSize)

        cropped = cropToAspectRatio(frame, self._colorSize)
        return cv2.resize(cropped, self._colorSize)

    def _createNewFrame(self, cvFrame) -> dai.ImgFrame:
        imgFrame = dai.ImgFrame()
        imgFrame.setData(cvFrame)
        imgFrame.setTimestamp(self._now)
        imgFrame.setSequenceNum(self._seqNum)
        shape = cvFrame.shape[::-1]
        imgFrame.setWidth(shape[0])
        imgFrame.setHeight(shape[1])
        return imgFrame

    def _createImgFrame(self, name: str, cvFrame) -> dai.ImgFrame:
        imgFrame: dai.ImgFrame
        if name == 'color':
            # Resize/crop color frame as specified by the user
            cvFrame = self._resizeColor(cvFrame)
            # cv2 reads frames in interleaved format, and most networks expect planar by default
            cvFrame = toPlanar(cvFrame)
            imgFrame = self._createNewFrame(cvFrame)
            imgFrame.setType(dai.RawImgFrame.Type.BGR888p)
            imgFrame.setInstanceNum(dai.CameraBoardSocket.RGB)
        elif name == 'left' or name == 'right':
            imgFrame = self._createNewFrame(cvFrame)
            imgFrame.setType(dai.RawImgFrame.Type.RAW8)
            imgFrame.setInstanceNum(getattr(dai.CameraBoardSocket, name.upper()))
        elif name == 'depth':
            imgFrame = self._createNewFrame(cvFrame)
            imgFrame.setType(dai.RawImgFrame.Type.RAW16)

        return imgFrame

    def _readFrames(self) -> bool:
        """
        Reads frames from all Readers.
        
        Returns:
            bool: True if successful, otherwise False.
        """
        self.frames = dict()
        frames = self.reader.read()
        if not frames:
            return False  # No more frames!

        for name, frame in frames.items():
            self.frames[name] = frame

        # Compress 3-plane frame to a single plane
        for name, frame in self.frames.items():
            if name in ["left", "right", "disparity"] and len(frame.shape) == 3:
                self.frames[name] = frame[:, :, 0]  # All 3 planes are the same
        return True

    def _getMaxSize(self, name: str) -> int:
        """
        Used when setting XLinkIn nodes, so they consume the least amount of memory needed.
        """
        size = self.getShape(name)
        bytes_per_pixel = 1
        if name == 'color':
            bytes_per_pixel = 3
        elif name == 'depth':
            bytes_per_pixel = 2  # 16bit
        return size[0] * size[1] * bytes_per_pixel

    def getShape(self, name: str) -> Tuple[int, int]:
        """
        Get shape of a stream
        """
        if name in self.reader.getStreams():
            return self.reader.getShape(name)

    def close(self):
        """
        Closes all video readers.
        """
        self._stop = True
        if self.thread:
            self.thread.join()
        self.reader.close()



import os
from abc import ABC, abstractmethod
from pathlib import Path
from typing import List, Tuple


class AbstractReader(ABC):
    @abstractmethod
    def read(self):
        """
        Read a frame (or multiple frames) from the reader.
        @return: Single np.ndarray, or dict of frames and their names. None if frames weren't read or there was an error.
        """
        pass

    @abstractmethod
    def getStreams(self) -> List[str]:
        pass

    @abstractmethod
    def getShape(self, name: str) -> Tuple[int, int]:
        """
        Returns (width, height)
        """
        pass

    @abstractmethod
    def close(self):
        pass

    @abstractmethod
    def disableStream(self, name: str):
        """
        @param name: Name of the stream to be disabled
        """
        pass

    def _fileWithExt(self, folder: Path, ext: str) -> str:
        for f in os.listdir(str(folder)):
            if f.endswith(ext):
                return f
        raise ValueError(f"Couldn't find a file with '{ext}' extension in folder '{folder}'!")



from Qt import QtWidgets, QtCore
from NodeGraphQt import NodeGraph, BaseNode, PropertiesBinWidget, Port
import time

# create a node class object inherited from BaseNode.
class FooNode(BaseNode):

    # unique node identifier domain.
    __identifier__ = 'com.chantasticvfx'

    # initial default node name.
    NODE_NAME = 'Foo Node'

    def __init__(self):
        super(FooNode, self).__init__()

        # create an input port.
        self.add_input('in')

        # create an output port.
        self.add_output('out')


if __name__ == '__main__':
    QtCore.QCoreApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling)
    app = QtWidgets.QApplication(["DepthAI Pipeline Graph"])

    # create node graph controller.
    graph = NodeGraph()

    # set up default menu and commands.
    # setup_context_menu(graph)

    # register the FooNode node class.
    graph.register_node(FooNode)

    # show the node graph widget.
    properties_bin = PropertiesBinWidget(node_graph=graph)

    properties_bin.setWindowFlags(QtCore.Qt.Tool)

        # show the node properties bin widget when a node is double clicked.
    def display_properties_bin(node):
        global properties_bin
        if not properties_bin.isVisible():
            properties_bin.show()

    # wire function to "node_double_clicked" signal.
    graph.node_double_clicked.connect(display_properties_bin)

    # create two nodes.
    node_a = graph.create_node('com.chantasticvfx.FooNode', name='node A')
    output = node_a.add_output(name='out1')
    node_b = graph.create_node('com.chantasticvfx.FooNode', name='node B', pos=(300, 50))
    input = node_b.add_input(name='in1')

    link = input.connect_to(output)

    graph.widget.show()


    while True:
        time.sleep(0.1)
        link.new_event(sender=True)
        link.new_event(sender=True)
        link.new_event(sender=False)
        # output.view.name = str(i)
        app.processEvents()

    app.exec_()


from typing import Dict, List
import requests
import depthai as dai
from pathlib import Path
import json
from zipfile import ZipFile

ROBOFLOW_MODELS = Path.home() / Path('.cache/roboflow-models')

class RoboflowIntegration:
    config: Dict

    def __init__(self, config: Dict):
        if 'key' not in config:
            raise ValueError("To download a model from Roboflow, you need to pass API key ('key')!")
        if 'model' not in config:
            raise ValueError("To download a model from Roboflow, you need to pass model path ('model')!")

        self.config = config

    def _file_with_ext(self, folder: Path, ext: str) -> Path:
        files = list(folder.glob(f"*{ext}"))
        if 0 == len(files):
            return None
        return files[0]

    def device_update(self, device: dai.Device) -> Path:
        mxid = device.getMxId()
        url = f"https://api.roboflow.com/depthai/{self.config['model']}/?api_key={self.config['key']}&device={mxid}"
        response = requests.get(url)

        name = self.config['model'].replace('/', '_') # '/' isn't valid folder name

        model_folder = ROBOFLOW_MODELS / name
        jsonFile = self._file_with_ext(model_folder, '.json')
        if jsonFile:
            return jsonFile

        json_res = response.json()
        if "error" in json_res:
            raise Exception(json_res['error'])

        response.raise_for_status()
        ret = json_res['depthai']

        if not str(ret['modelType']).startswith('yolov'):
            raise ValueError("This Roboflow's model is not from YOLO family!")

        if not str(ret['modelType']).endswith('n'):
            print('We recommend using a lighter version of the model to get a better performance!')

        print(f"Downloading '{ret['name']}' model from Roboflow server")

        zipFileReq = requests.get(ret['model'])
        zipFileReq.raise_for_status()

        (ROBOFLOW_MODELS / name).mkdir(parents=True, exist_ok=True)
        zipFilePath = str(ROBOFLOW_MODELS / 'roboflow.zip')
        # Download the .zip where our model is
        with open(zipFilePath, 'wb') as f:
            f.write(zipFileReq.content)

        print(f"Downloaded the model to {zipFilePath}")

        with ZipFile(zipFilePath, 'r') as zObject: # Extract the zip
            zObject.extractall(str(ROBOFLOW_MODELS / name))

        # Rename bin/xml files
        self._file_with_ext(model_folder, ".xml").rename(str(model_folder / (name + ".xml")))
        self._file_with_ext(model_folder, ".bin").rename(str(model_folder / (name + ".bin")))

        # Rename bin/xml paths inside the json
        new_json_name = str(model_folder / (name + ".json"))
        jsonConf = json.load(self._file_with_ext(model_folder, ".json").open())
        jsonConf['model']['xml'] = name + ".xml"
        jsonConf['model']['bin'] = name + ".bin"

        # For some reason, Roboflow server provides incorrect json file, so we have to edit it
        if 'output_format' in jsonConf:
            jsonConf['nn_config']['output_format'] = jsonConf['output_format']
            del jsonConf['output_format']
        if 'NN_family' in jsonConf:
            jsonConf['nn_config']['NN_family'] = jsonConf['NN_family']
            del jsonConf['NN_family']

        # Overwrite the json file, and rename it
        self._file_with_ext(model_folder, ".json").rename(new_json_name).open("w").write(json.dumps(jsonConf))

        Path(zipFilePath).unlink()  # Delete .zip

        return Path(new_json_name)



import dataclasses
import json

import numpy as np


class JSONEncoder(json.JSONEncoder):
    """ Special json encoder for numpy types """

    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif dataclasses.is_dataclass(obj):
            return dataclasses.asdict(obj)

        return json.JSONEncoder.default(self, obj)



import cv2
import numpy as np
from depthai import NNData

from depthai_sdk.classes.nn_results import ImgLandmarks

keypoints_mapping = [
    'Nose', 'Neck', 'R-Sho', 'R-Elb', 'R-Wr',
    'L-Sho', 'L-Elb', 'L-Wr', 'R-Hip', 'R-Knee',
    'R-Ank', 'L-Hip', 'L-Knee', 'L-Ank', 'R-Eye',
    'L-Eye', 'R-Ear', 'L-Ear'
]
POSE_PAIRS = [
    [1, 2], [1, 5], [2, 3], [3, 4], [5, 6], [6, 7], [1, 8], [8, 9], [9, 10], [1, 11],
    [11, 12], [12, 13], [1, 0], [0, 14], [14, 16], [0, 15], [15, 17], [2, 17], [5, 16]
]
map_idx = [
    [31, 32], [39, 40], [33, 34], [35, 36], [41, 42], [43, 44], [19, 20], [21, 22], [23, 24], [25, 26],
    [27, 28], [29, 30], [47, 48], [49, 50], [53, 54], [51, 52], [55, 56], [37, 38], [45, 46]
]
colors = [
    [0, 100, 255], [0, 100, 255], [0, 255, 255], [0, 100, 255], [0, 255, 255], [0, 100, 255], [0, 255, 0],
    [255, 200, 100], [255, 0, 255], [0, 255, 0], [255, 200, 100], [255, 0, 255], [0, 0, 255], [255, 0, 0],
    [200, 200, 0], [255, 0, 0], [200, 200, 0], [0, 0, 0]
]

NN_HEIGHT, NN_WIDTH = 256, 456

threshold = 0.3
n_points = 18


def decode(nn_data: NNData) -> ImgLandmarks:
    heatmaps = np.array(nn_data.getLayerFp16('Mconv7_stage2_L2')).reshape((1, 19, 32, 57))
    pafs = np.array(nn_data.getLayerFp16('Mconv7_stage2_L1')).reshape((1, 38, 32, 57))
    heatmaps = heatmaps.astype('float32')
    pafs = pafs.astype('float32')
    outputs = np.concatenate((heatmaps, pafs), axis=1)

    new_keypoints = []
    new_keypoints_list = np.zeros((0, 3))
    keypoint_id = 0

    for row in range(n_points):
        prob_map = outputs[0, row, :, :]
        prob_map = cv2.resize(prob_map, (NN_WIDTH, NN_HEIGHT))  # (456, 256)
        keypoints = get_keypoints(prob_map, threshold=threshold)
        new_keypoints_list = np.vstack([new_keypoints_list, *keypoints])
        keypoints_with_id = []

        for i in range(len(keypoints)):
            keypoints_with_id.append(keypoints[i] + (keypoint_id,))
            keypoint_id += 1

        new_keypoints.append(keypoints_with_id)

    valid_pairs, invalid_pairs = get_valid_pairs(outputs, w=NN_WIDTH, h=NN_HEIGHT, detected_keypoints=new_keypoints)
    new_personwise_keypoints = get_personwise_keypoints(valid_pairs, invalid_pairs, new_keypoints_list)

    keypoint_points = []

    for n in range(len(new_personwise_keypoints)):
        person_keypoints = []
        for i in range(n_points - 1):
            index = new_personwise_keypoints[n][np.array(POSE_PAIRS[i])]

            if -1 in index:
                continue

            k1 = np.int32(new_keypoints_list[index.astype(int), 0]) / NN_WIDTH
            k2 = np.int32(new_keypoints_list[index.astype(int), 1]) / NN_HEIGHT
            person_keypoints.append([[k1[0], k2[0]], [k1[1], k2[1]]])

        keypoint_points.append(person_keypoints)

    return ImgLandmarks(nn_data, keypoint_points, POSE_PAIRS, colors)


def get_keypoints(prob_map, threshold=0.2):
    map_smooth = cv2.GaussianBlur(prob_map, (3, 3), 0, 0)
    map_mask = np.uint8(map_smooth > threshold)
    keypoints = []

    try:
        # OpenCV4.x
        contours, _ = cv2.findContours(map_mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
    except:
        # OpenCV3.x
        _, contours, _ = cv2.findContours(map_mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)

    for cnt in contours:
        blob_mask = np.zeros(map_mask.shape)
        blob_mask = cv2.fillConvexPoly(blob_mask, cnt, 1)
        masked_prob_map = map_smooth * blob_mask
        _, max_val, _, max_loc = cv2.minMaxLoc(masked_prob_map)
        keypoints.append(max_loc + (prob_map[max_loc[1], max_loc[0]],))

    return keypoints


def get_valid_pairs(outputs, w, h, detected_keypoints):
    valid_pairs = []
    invalid_pairs = []
    n_interp_samples = 10
    paf_score_th = 0.2
    conf_th = 0.4

    for k in range(len(map_idx)):
        paf_a = outputs[0, map_idx[k][0], :, :]
        paf_b = outputs[0, map_idx[k][1], :, :]
        paf_a = cv2.resize(paf_a, (w, h))
        paf_b = cv2.resize(paf_b, (w, h))

        cand_a = detected_keypoints[POSE_PAIRS[k][0]]
        cand_b = detected_keypoints[POSE_PAIRS[k][1]]

        n_a = len(cand_a)
        n_b = len(cand_b)

        if (n_a != 0 and n_b != 0):
            valid_pair = np.zeros((0, 3))
            for i in range(n_a):
                max_j = -1
                max_score = -1
                found = 0
                for j in range(n_b):
                    d_ij = np.subtract(cand_b[j][:2], cand_a[i][:2])
                    norm = np.linalg.norm(d_ij)
                    if norm:
                        d_ij = d_ij / norm
                    else:
                        continue
                    interp_coord = list(zip(np.linspace(cand_a[i][0], cand_b[j][0], num=n_interp_samples),
                                            np.linspace(cand_a[i][1], cand_b[j][1], num=n_interp_samples)))
                    paf_interp = []
                    for k in range(len(interp_coord)):
                        paf_interp.append([paf_a[int(round(interp_coord[k][1])), int(round(interp_coord[k][0]))],
                                           paf_b[int(round(interp_coord[k][1])), int(round(interp_coord[k][0]))]])
                    paf_scores = np.dot(paf_interp, d_ij)
                    avg_paf_score = sum(paf_scores) / len(paf_scores)

                    if (len(np.where(paf_scores > paf_score_th)[0]) / n_interp_samples) > conf_th:
                        if avg_paf_score > max_score:
                            max_j = j
                            max_score = avg_paf_score
                            found = 1
                if found:
                    valid_pair = np.append(valid_pair, [[cand_a[i][3], cand_b[max_j][3], max_score]], axis=0)

            valid_pairs.append(valid_pair)
        else:
            invalid_pairs.append(k)
            valid_pairs.append([])

    return valid_pairs, invalid_pairs


def get_personwise_keypoints(valid_pairs, invalid_pairs, keypoints_list):
    personwise_keypoints = -1 * np.ones((0, 19))

    for k in range(len(map_idx)):
        if k in invalid_pairs:
            continue

        part_as = valid_pairs[k][:, 0]
        part_bs = valid_pairs[k][:, 1]
        index_a, index_b = np.array(POSE_PAIRS[k])

        for i in range(len(valid_pairs[k])):
            found = 0
            person_idx = -1
            for j in range(len(personwise_keypoints)):
                if personwise_keypoints[j][index_a] == part_as[i]:
                    person_idx = j
                    found = 1
                    break

            if found:
                personwise_keypoints[person_idx][index_b] = part_bs[i]
                personwise_keypoints[person_idx][-1] += keypoints_list[part_bs[i].astype(int), 2] + \
                                                        valid_pairs[k][i][2]

            elif not found and k < 17:
                row = -1 * np.ones(19)
                row[index_a] = part_as[i]
                row[index_b] = part_bs[i]
                row[-1] = sum(keypoints_list[valid_pairs[k][i, :2].astype(int), 2]) + valid_pairs[k][i][2]
                personwise_keypoints = np.vstack([personwise_keypoints, row])

    return personwise_keypoints



import os

from Qt import QtWidgets

_current_user_directory = os.path.expanduser('~')


def set_dir(file):
    global _current_user_directory
    if os.path.isdir(file):
        _current_user_directory = file
    elif os.path.isfile(file):
        _current_user_directory = os.path.split(file)[0]


class FileDialog(object):

    @staticmethod
    def getSaveFileName(parent=None, title='Save File', file_dir=None,
                        ext_filter='*'):
        if not file_dir:
            file_dir = _current_user_directory
        file_dlg = QtWidgets.QFileDialog.getSaveFileName(
            parent, title, file_dir, ext_filter)
        file = file_dlg[0] or None
        if file:
            set_dir(file)
        return file_dlg

    @staticmethod
    def getOpenFileName(parent=None, title='Open File', file_dir=None,
                        ext_filter='*'):
        if not file_dir:
            file_dir = _current_user_directory
        file_dlg = QtWidgets.QFileDialog.getOpenFileName(
            parent, title, file_dir, ext_filter)
        file = file_dlg[0] or None
        if file:
            set_dir(file)
        return file_dlg


class BaseDialog(object):

    @staticmethod
    def message_dialog(text='', title='Message'):
        dlg = QtWidgets.QMessageBox()
        dlg.setWindowTitle(title)
        dlg.setInformativeText(text)
        dlg.setStandardButtons(QtWidgets.QMessageBox.Ok)
        return dlg.exec()

    @staticmethod
    def question_dialog(text='', title='Are you sure?'):
        dlg = QtWidgets.QMessageBox()
        dlg.setWindowTitle(title)
        dlg.setInformativeText(text)
        dlg.setStandardButtons(
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No
        )
        result = dlg.exec()
        return bool(result == QtWidgets.QMessageBox.Yes)



import traceback
from pathlib import Path
from ..previews import Previews
import depthai as dai
import os

class EncodingManager:
    """
    Manager class that handles video encoding
    """

    _encodingQueues = {}
    _encodingNodes = {}
    _encodingFiles = {}

    def __init__(self, encodeConfig: dict, encodeOutput=None):
        """
        Args:
            encodeConfig (dict): Encoding config consisting of keys as preview names and values being the encoding FPS
            encodeOutput (pathlib.Path, Optional): Output directory for the recorded videos
        """
        self.encodeConfig = encodeConfig
        self.encodeOutput = Path(encodeOutput) if encodeOutput is not None else None
        if self.encodeOutput is not None and not self.encodeOutput.exists():
            os.makedirs(self.encodeOutput, exist_ok=True)

    def createEncoders(self, pm):
        """
        Creates VideoEncoder nodes using Pipeline manager, based on config provided during initialization

        Args:
            pm (depthai_sdk.managers.PipelineManager): Pipeline Manager instance
        """

        self._encodingNodes.clear()
        for cameraName, encFps in self.encodeConfig.items():
            pm.createEncoder(cameraName, encFps)
            self._encodingNodes[cameraName] = getattr(pm.nodes, cameraName + "Enc")

    def createDefaultQueues(self, device):
        """
        Creates output queues for VideoEncoder nodes created in :code:`create_encoders` function. Also, opems up
        the H.264 / H.265 stream files (e.g. :code:`color.h265`) where the encoded data will be stored.

        Args:
            device (depthai.Device): Running device instance
        """
        self._encodingQueues.clear()
        self._encodingFiles.clear()

        for cameraName, node in self._encodingNodes.items():
            self._encodingQueues[cameraName] = device.getOutputQueue(cameraName + "EncXout", maxSize=30, blocking=True)
            suffix = ".h265" if node.getProfile() == dai.VideoEncoderProperties.Profile.H265_MAIN else ".h264"
            self._encodingFiles[cameraName] = (self.encodeOutput / cameraName).with_suffix(suffix).open('wb')

    def parseQueues(self):
        """
        Parse the output queues, consuming the available data packets in it and storing them inside opened stream files
        """
        for name, queue in self._encodingQueues.items():
            while queue.has():
                queue.get().getData().tofile(self._encodingFiles[name])

    def close(self):
        """
        Closes opened stream files and tries to perform FFMPEG-based conversion from raw stream into mp4 video.

        If successful, each stream file (e.g. :code:`color.h265`) will be available along with a ready to use video file
        (e.g. :code:`color.mp4`).

        In case of failure, this method will print traceback and commands that can be used for manual conversion
        """

        def printManual():
            print(
                "To view the encoded data, convert the stream file (.h264/.h265) into a video file (.mp4), using commands below:")
            cmd = "ffmpeg -framerate {} -i {} -c copy {}"
            for name, file in self._encodingFiles.items():
                print(cmd.format(self._encodingNodes[name].getFrameRate(), file.name, str(Path(file.name).with_suffix('.mp4'))))

        for queue in self._encodingQueues.values():
            queue.close()

        for name, file in self._encodingFiles.items():
            file.close()
        try:
            import ffmpy3
            for name, file in self._encodingFiles.items():
                fps = self._encodingNodes[name].getFrameRate()
                outName = str(Path(file.name).with_suffix('.mp4'))
                try:
                    ff = ffmpy3.FFmpeg(
                        inputs={file.name: "-y"},
                        outputs={outName: "-c copy -framerate {}".format(fps)}
                    )
                    print("Running conversion command... [{}]".format(ff.cmd))
                    ff.run()
                except ffmpy3.FFExecutableNotFoundError:
                    print("FFMPEG executable not found!")
                    traceback.print_exc()
                    printManual()
                except ffmpy3.FFRuntimeError:
                    print("FFMPEG runtime error!")
                    traceback.print_exc()
                    printManual()
            print("Video conversion complete!")
            for name, file in self._encodingFiles.items():
                print("Produced file: {}".format(str(Path(file.name).with_suffix('.mp4'))))
        except ImportError:
            print("Module ffmpy3 not fouund!")
            traceback.print_exc()
            printManual()
        except:
            print("Unknown error!")
            traceback.print_exc()
            printManual()



import cv2
import numpy as np

from depthai_sdk import Previews, toTensorResult


def decode(nnManager, packet):
    # [print(f"Layer name: {l.name}, Type: {l.dataType}, Dimensions: {l.dims}") for l in packet.getAllLayers()]
    # after squeeze the data.shape is 4,512, 896
    data = np.squeeze(toTensorResult(packet)["L0317_ReWeight_SoftMax"])
    classColors = [[0, 0, 0], [0, 255, 0], [255, 0, 0], [0, 0, 255]]
    classColors = np.asarray(classColors, dtype=np.uint8)

    indices = np.argmax(data, axis=0)
    outputColors = np.take(classColors, indices, axis=0)
    return outputColors


def draw(nnManager, data, frames):
    if len(data) == 0:
        return
    for name, frame in frames:
        if name == "color" and nnManager.source == "color" and not nnManager._fullFov:
            scaleFactor = frame.shape[0] / nnManager.inputSize[1]
            resizeW = int(nnManager.inputSize[0] * scaleFactor)
            resized = cv2.resize(data, (resizeW, frame.shape[0])).astype(data.dtype)
            offsetW = int(frame.shape[1] - nnManager.inputSize[0] * scaleFactor) // 2
            tailW = frame.shape[1] - offsetW - resizeW
            stacked = np.hstack((np.zeros((frame.shape[0], offsetW, 3)).astype(resized.dtype), resized, np.zeros((frame.shape[0], tailW, 3)).astype(resized.dtype)))
            cv2.addWeighted(frame, 1, stacked, 0.2, 0, frame)
        elif name in (Previews.color.name, Previews.nnInput.name, "host"):
            cv2.addWeighted(frame, 1, cv2.resize(data, frame.shape[:2][::-1]), 0.2, 0, frame)



from pathlib import Path
from typing import List, Tuple

import cv2
import numpy as np
from mcap.mcap0.reader import make_reader
from mcap.mcap0.stream_reader import StreamReader
from mcap_ros1.decoder import Decoder

from depthai_sdk.previews import PreviewDecoder
from depthai_sdk.readers.abstract_reader import AbstractReader


class McapReader(AbstractReader):
    """
    Reads all saved streams from .mcap recording.
    Supported ROS messages: Image (depth), CompressedImage (left, right, color, disparity)
    """
    _readFrames = dict()

    def __init__(self, folder: Path) -> None:
        # Get available topics
        with open(source, "rb") as file:
            reader = make_reader(file)
            channels = reader.get_summary().channels
            self._topics = [c.topic.split('/')[0] for _, c in channels.items()]

        # Init msg array
        for topic in self._topics:
            self._readFrames[topic] = []

        # Create MCAP decoder
        decoder = Decoder(StreamReader(str(source)))
        self.msgs = decoder.messages
        # Prepare initial frames so we can get frame size
        self._prepareFrames()

    def read(self):
        """
        Read and return one frame from each available stream.
        """
        self._prepareFrames()
        return self._returnFrames()

    def _prepareFrames(self):
        """
        Read frames until one of each is buffered. Afterwards, call self._returnFrames()
        to get one of each buffered frame.
        """
        while not self._framesReady():
            topic, _, msg = next(self.msgs)
            name = topic.split('/')[0]
            self._readFrames[name].append(self._getCvFrame(msg, name))

    def _getCvFrame(self, msg, name: str):
        """
        Convert ROS message to cv2 frame (numpy array)
        """
        msgType = str(type(msg))
        data = np.frombuffer(msg.data, dtype=np.int8)
        if 'CompressedImage' in msgType:
            if name == 'color':
                return PreviewDecoder.jpegDecode(data, cv2.IMREAD_COLOR)
            else:  # left, right, disparity
                return PreviewDecoder.jpegDecode(data, cv2.IMREAD_GRAYSCALE)
        elif 'Image' in msgType:
            if msg.encoding == 'mono16':
                data = data.view(np.int16)
            return data.reshape((msg.height, msg.width))
            # msg.encoding
        else:
            raise Exception('Only CompressedImage and Image ROS messages are currently supported.')

    def _framesReady(self):
        """
        Check if there is at least one frame from each available stream.
        """
        for _, arr in self._readFrames.items():
            if len(arr) < 1: return False
        return True

    def _returnFrames(self):
        """
        Return synced frames (one of each available).
        """
        ret = dict()
        for name, arr in self._readFrames.items():
            ret[name] = arr.pop(0)
        return ret

    def getStreams(self) -> List[str]:
        """
        Get available topics
        """
        return [name for name in self._topics]

    def getShape(self, name: str) -> Tuple[int, int]:
        frame = self._readFrames[name][0]
        return (frame.shape[1], frame.shape[0])

    def close(self):
        pass



from collections import deque
from datetime import datetime
from pathlib import Path
from typing import Union

import cv2
import depthai as dai
import numpy as np

from depthai_sdk.recorders.video_writers import AbstractWriter
from depthai_sdk.recorders.video_writers.utils import create_writer_dir


class VideoWriter(AbstractWriter):
    _fps: float
    _path: str

    def __init__(self, path: Path, name: str, fourcc: str, fps: float):
        self.file = None
        self._path = create_writer_dir(path, name, 'avi')
        self._fourcc = None

        self._w, self._h = None, None
        self._fps = fps

        self._buffer = None
        self._is_buffer_enabled = False

    def init_buffer(self, max_seconds: int):
        if max_seconds > 0:
            self._buffer = deque(maxlen=int(max_seconds * self._fps))
            self._is_buffer_enabled = True

    def _create_file(self, frame: Union[dai.ImgFrame, np.ndarray]):
        if isinstance(frame, np.ndarray):
            self._h, self._w = frame.shape[:2]
        else:
            self._h, self._w = frame.getHeight(), frame.getWidth()

        # Disparity - RAW8
        # Depth - RAW16
        if self._fourcc is None:
            if isinstance(frame, np.ndarray):
                c = 1 if frame.ndim == 2 else frame.shape[2]
                self._fourcc = "GRAY" if c == 1 else "I420"
            else:
                if frame.getType() == dai.ImgFrame.Type.RAW16:  # Depth
                    self._fourcc = "FFV1"
                elif frame.getType() == dai.ImgFrame.Type.RAW8:  # Mono Cams
                    self._fourcc = "GREY"
                else:
                    self._fourcc = "I420"

        self.file = cv2.VideoWriter(self._path,
                                    cv2.VideoWriter_fourcc(*self._fourcc),
                                    self._fps,
                                    (self._w, self._h),
                                    isColor=self._fourcc != "GREY")

    def close(self):
        if self.file:
            self.file.release()

    def save_snapshot(self, duration: int, dir_path: Union[Path, str] = None):
        if self._buffer is None:
            raise RuntimeError("Buffer is not enabled")

        if len(self._buffer) == 0:
            return None

        snapshot_name = f'snapshot_{datetime.now().strftime("%Y-%m-%d_%H-%M-%S")}.avi'
        save_path = Path(dir_path or self._path.partition("/")[0], snapshot_name)

        snapshot_file = cv2.VideoWriter(
            str(save_path),
            cv2.VideoWriter_fourcc(*self._fourcc),
            self._fps,
            (self._w, self._h),
            isColor=self._fourcc == "I420"
        )

        # Copy queue
        buffer_copy = self._buffer.copy()

        n_skip_frames = int(self._fps * (self._fps * duration))
        while len(buffer_copy) > 0:
            # Wait til we reach the desired time
            if n_skip_frames > 0:
                n_skip_frames -= 1
                buffer_copy.popleft()
                continue

            el = buffer_copy.popleft()
            snapshot_file.write(el if isinstance(el, np.ndarray) else el.getCvFrame())

        snapshot_file.release()
        print('Snapshot saved to', save_path)

    def set_fourcc(self, fourcc: str):
        self._fourcc = fourcc

    def add_to_buffer(self, frame: Union[dai.ImgFrame, np.ndarray]):
        if not self._is_buffer_enabled:
            return

        if len(self._buffer) == self._buffer.maxlen:
            self._buffer.pop()

        self._buffer.append(frame)

    def write(self, frame: Union[dai.ImgFrame, np.ndarray]):
        if self.file is None:
            self._create_file(frame)

        self.add_to_buffer(frame)

        self.file.write(frame if isinstance(frame, np.ndarray) else frame.getCvFrame())



from typing import Dict

from depthai_sdk.components.camera_helper import *
from depthai_sdk.components.component import Component
from depthai_sdk.components.parser import parse_resolution, parse_encode
from depthai_sdk.oak_outputs.xout import XoutFrames, XoutMjpeg, XoutH26x
from depthai_sdk.oak_outputs.xout_base import XoutBase, StreamXout, ReplayStream
from depthai_sdk.replay import Replay


class CameraComponent(Component):
    # Users should have access to these nodes
    node: Union[dai.node.MonoCamera, dai.node.XLinkIn, dai.node.ColorCamera]
    encoder: dai.node.VideoEncoder

    stream: dai.Node.Output  # Node output to be used as eg. an input into NN
    stream_size: Tuple[int, int]  # Output size

    # Setting passed at init
    _replay: Replay  # Replay module
    _args: Dict
    _control: bool
    _source: str

    # Parsed from settings
    _encoderProfile: dai.VideoEncoderProperties.Profile = None

    def __init__(self,
                 pipeline: dai.Pipeline,
                 source: str,
                 resolution: Union[
                     None, str, dai.ColorCameraProperties.SensorResolution, dai.MonoCameraProperties.SensorResolution] = None,
                 fps: Optional[float] = None,
                 encode: Union[None, str, bool, dai.VideoEncoderProperties.Profile] = None,
                 rotation: Optional[int] = None,
                 replay: Optional[Replay] = None,
                 args: Dict = None,
                 ):
        """
        Creates Camera component. This abstracts ColorCamera/MonoCamera nodes and supports mocking the camera when
        recording is passed during OakCamera initialization. Mocking the camera will send frames from the host to the
        OAK device (via XLinkIn node).

        Args:
            source (str): Source of the camera. Either color/rgb/right/left
            resolution (optional): Camera resolution, eg. '800p' or '4k'
            fps (float, optional): Camera FPS
            out (bool, default False): Whether we want to stream frames to the host computer
            encode: Encode streams before sending them to the host. Either True (use default), or mjpeg/h264/h265
            replay (Replay object): Replay
            args (Dict): Use user defined arguments when constructing the pipeline
        """
        super().__init__()
        self.out = self.Out(self)

        # Save passed settings
        self._source = source
        self._replay: Replay = replay
        self._args = args

        if rotation not in [None, 0, 90, 180, 270]:
            raise ValueError(f'Angle {rotation} not supported! Use 0, 90, 180, 270.')

        self._rotation = rotation

        self._create_node(pipeline, source.upper())

        self.encoder = None
        if encode:
            self.encoder = pipeline.createVideoEncoder()
            # MJPEG by default
            self._encoderProfile = parse_encode(encode)

        self._resolution_forced: bool = resolution is not None
        if resolution:
            self._set_resolution(resolution)
        if fps:
            self._set_fps(fps)

    def _update_device_info(self, pipeline: dai.Pipeline, device: dai.Device, version: dai.OpenVINO.Version):
        if self.is_replay():  # If replay, don't create camera node
            res = self._replay.getShape(self._source)
            # print('resolution', res)
            # resize = getResize(res, width=1200)
            # self._replay.setResizeColor(resize)
            self.node: dai.node.XLinkIn = getattr(self._replay, self._source)
            # print('resize', resize)
            self.node.setMaxDataSize(res[0] * res[1] * 3)
            self.stream_size = res
            self.stream = self.node.out
            if self._rotation:
                rot_manip = self._create_rotation_manip(pipeline) if self._rotation else None
                self.node.out.link(rot_manip.inputImage)
                self.stream = rot_manip.out
            else:
                self.stream = self.node.out

        if isinstance(self.node, dai.node.ColorCamera):
            # DepthAI uses CHW (Planar) channel layout convention for NN inferencing
            self.node.setInterleaved(False)
            # DepthAI uses BGR color order convention for NN inferencing
            self.node.setColorOrder(dai.ColorCameraProperties.ColorOrder.BGR)
            self.node.setPreviewNumFramesPool(4)

            cams = device.getCameraSensorNames()
            # print('Available sensors on OAK:', cams)
            sensor_name = cams[dai.CameraBoardSocket.RGB]

            if not self._resolution_forced:  # Find the closest resolution
                self.node.setResolution(getClosesResolution(sensor_name, width=1200))
                scale = getClosestIspScale(self.node.getIspSize(), width=1200, videoEncoder=(self.encoder is not None))
                self.node.setIspScale(*scale)

            self.node.setVideoSize(*getClosestVideoSize(*self.node.getIspSize()))
            self.node.setVideoNumFramesPool(2)  # We will increase it later if we are streaming to host

            self.node.setPreviewSize(*self.node.getVideoSize())
            self.stream_size = self.node.getPreviewSize()
            self.stream = self.node.preview if self.encoder is None else self.node.video

        elif isinstance(self.node, dai.node.MonoCamera):
            self.stream_size = self.node.getResolutionSize()
            self.stream = self.node.out

        if self._args:
            self._config_camera_args(self._args)

        if self._rotation:
            rot_manip = self._create_rotation_manip(pipeline) if self._rotation else None
            self.stream.link(rot_manip.inputImage)
            self.stream = rot_manip.out

        if self.encoder:
            self.encoder.setDefaultProfilePreset(self._get_fps(), self._encoderProfile)
            if self.is_replay():  # TODO - this might be not needed, we check for replay above and return
                # Create ImageManip to convert to NV12
                type_manip = pipeline.createImageManip()
                type_manip.setFrameType(dai.ImgFrame.Type.NV12)
                type_manip.setMaxOutputFrameSize(self.stream_size[0] * self.stream_size[1] * 3)

                self.stream.link(type_manip.inputImage)
                type_manip.out.link(self.encoder.input)
            elif self.is_mono():
                self.stream.link(self.encoder.input)
            elif self.is_color():
                self.stream.link(self.encoder.input)
            else:
                raise ValueError('CameraComponent is neither Color, Mono, nor Replay!')

    def _create_rotation_manip(self, pipeline: dai.Pipeline):
        rot_manip = pipeline.createImageManip()
        rgb_rr = dai.RotatedRect()
        w, h = self.stream_size
        rgb_rr.center.x, rgb_rr.center.y = w // 2, h // 2
        rgb_rr.size.width, rgb_rr.size.height = (w, h) if self._rotation % 180 == 0 else (h, w)
        rgb_rr.angle = self._rotation
        rot_manip.initialConfig.setCropRotatedRect(rgb_rr, False)
        rot_manip.setMaxOutputFrameSize(w * h * 3)
        return rot_manip

    def _create_node(self, pipeline: dai.Pipeline, source: str) -> None:
        """
        Called from __init__ to parse passed `source` argument.
        @param source: User-input source
        """
        if source == "COLOR" or source == "RGB":
            if self.is_replay():
                if 'color' not in self._replay.getStreams():
                    raise Exception('Color stream was not found in specified depthai-recording!')
            else:
                self.node = pipeline.createColorCamera()
                self.node.setBoardSocket(dai.CameraBoardSocket.RGB)

        elif source == "RIGHT" or source == "MONO":
            if self.is_replay():
                if 'right' not in self._replay.getStreams():
                    raise Exception('Right stream was not found in specified depthai-recording!')
            else:
                self.node = pipeline.createMonoCamera()
                self.node.setBoardSocket(dai.CameraBoardSocket.RIGHT)

        elif source == "LEFT":
            if self.is_replay():
                if 'left' not in self._replay.getStreams():
                    raise Exception('Left stream was not found in specified depthai-recording!')
            else:
                self.node = pipeline.createMonoCamera()
                self.node.setBoardSocket(dai.CameraBoardSocket.LEFT)
        else:
            raise ValueError(f"Source name '{source}' not supported!")

    # Should be mono/color camera agnostic. Also call this from __init__ if args is enabled
    def config_camera(self,
                      preview: Union[None, str, Tuple[int, int]] = None,
                      fps: Optional[float] = None,
                      resolution: Union[
                          None, str, dai.ColorCameraProperties.SensorResolution, dai.MonoCameraProperties.SensorResolution] = None
                      ) -> None:
        """
        Configure resolution, scale, FPS, etc.
        """

        # TODO
        if fps: self._set_fps(fps)
        if resolution: self._set_resolution(resolution)

        if preview:
            from .parser import parse_size
            preview = parse_size(preview)

            self.stream_size = preview

            if self._replay:
                self._replay.setResizeColor(preview)
            elif self.is_color():
                self.node.setPreviewSize(preview)
            else:
                # TODO: Use ImageManip to set mono frame size

                raise NotImplementedError("Not yet implemented")

    def _config_camera_args(self, args: Dict):
        if not isinstance(args, Dict):
            args = vars(args)  # Namespace -> Dict

        if self.is_color():
            self.config_camera(
                fps=args.get('rgbFps', None),
                resolution=args.get('rgbResolution', None),
            )
            self.config_color_camera(
                manulFocus=args.get('manualFocus', None),
                afMode=args.get('afMode', None),
                awbMode=args.get('awbMode', None),
                sceneMode=args.get('sceneMode', None),
                antiBandingMode=args.get('antiBandingMode', None),
                effectMode=args.get('effectMode', None),
                ispScale=args.get('ispScale', None),
                sharpness=args.get('sharpness', None),
                lumaDenoise=args.get('lumaDenoise', None),
                chromaDenoise=args.get('chromaDenoise', None),
            )
        elif self.is_mono():
            self.config_camera(
                fps=args.get('monoFps', None),
                resolution=args.get('monoResolution', None),
            )
        else:  # Replay
            self.config_camera(fps=args.get('fps', None))

    def config_color_camera(self,
                            interleaved: Optional[bool] = None,
                            colorOrder: Union[None, dai.ColorCameraProperties.ColorOrder, str] = None,
                            # Cam control
                            manulFocus: Optional[int] = None,
                            afMode: Optional[dai.CameraControl.AutoFocusMode] = None,
                            awbMode: Optional[dai.CameraControl.AutoWhiteBalanceMode] = None,
                            sceneMode: Optional[dai.CameraControl.SceneMode] = None,
                            antiBandingMode: Optional[dai.CameraControl.AntiBandingMode] = None,
                            effectMode: Optional[dai.CameraControl.EffectMode] = None,
                            # IQ settings
                            ispScale: Optional[Tuple[int, int]] = None,
                            sharpness: Optional[int] = None,
                            lumaDenoise: Optional[int] = None,
                            chromaDenoise: Optional[int] = None,
                            ) -> None:
        if not self.is_color():
            print("Attempted to configure ColorCamera, but this component doesn't have it. Config attempt ignored.")
            return

        if self._replay is not None:
            print('Tried configuring ColorCamera, but replaying is enabled. Config attempt ignored.')
            return

        self.node: dai.node.ColorCamera

        if interleaved: self.node.setInterleaved(interleaved)
        if colorOrder:
            if isinstance(colorOrder, str):
                colorOrder = getattr(dai.ColorCameraProperties.ColorOrder, colorOrder.upper())
            self.node.setColorOrder(colorOrder)

        if manulFocus: self.node.initialControl.setManualFocus(manulFocus)
        if afMode: self.node.initialControl.setAutoFocusMode(afMode)
        if awbMode: self.node.initialControl.setAutoWhiteBalanceMode(awbMode)
        if sceneMode: self.node.initialControl.setSceneMode(sceneMode)
        if antiBandingMode: self.node.initialControl.setAntiBandingMode(antiBandingMode)
        if effectMode: self.node.initialControl.setEffectMode(effectMode)
        # EQ settings
        if ispScale:
            self._resolution_forced = True
            self.node.setIspScale(*ispScale)
        if sharpness: self.node.initialControl.setSharpness(sharpness)
        if lumaDenoise: self.node.initialControl.setLumaDenoise(lumaDenoise)
        if chromaDenoise: self.node.initialControl.setChromaDenoise(chromaDenoise)

    def _set_resolution(self, resolution):
        if not self.is_replay():
            self.node.setResolution(parse_resolution(type(self.node), resolution))
        # TODO: support potentially downscaling depthai-recording

    def is_replay(self) -> bool:
        return self._replay is not None

    def is_color(self) -> bool:
        return isinstance(self.node, dai.node.ColorCamera)

    def is_mono(self) -> bool:
        return isinstance(self.node, dai.node.MonoCamera)

    def _get_fps(self):
        return (self._replay if self.is_replay() else self.node).getFps()

    def _set_fps(self, fps: float):
        (self._replay if self._replay else self.node).setFps(fps)

    def config_encoder_h26x(self,
                            rate_control_mode: Optional[dai.VideoEncoderProperties.RateControlMode] = None,
                            keyframe_freq: Optional[int] = None,
                            bitrate_kbps: Optional[int] = None,
                            num_b_frames: Optional[int] = None,
                            ):
        if self.encoder is None:
            raise Exception('Video encoder was not enabled!')
        if self._encoderProfile == dai.VideoEncoderProperties.Profile.MJPEG:
            raise Exception('Video encoder was set to MJPEG while trying to configure H26X attributes!')

        if rate_control_mode:
            self.encoder.setRateControlMode(rate_control_mode)
        if keyframe_freq:
            self.encoder.setKeyframeFrequency(keyframe_freq)
        if bitrate_kbps:
            self.encoder.setBitrateKbps(bitrate_kbps)
        if num_b_frames:
            self.encoder.setNumBFrames(num_b_frames)

    def config_encoder_mjpeg(self,
                             quality: Optional[int] = None,
                             lossless: bool = False
                             ):
        if self.encoder is None:
            raise Exception('Video encoder was not enabled!')
        if self._encoderProfile != dai.VideoEncoderProperties.Profile.MJPEG:
            raise Exception(
                f'Video encoder was set to {self._encoderProfile} while trying to configure MJPEG attributes!')

        if quality:
            self.encoder.setQuality(quality)
        if lossless:
            self.encoder.setLossless(lossless)

    def get_stream_xout(self) -> StreamXout:
        if self.is_replay():
            return ReplayStream(self._source)
        elif self.is_mono():
            return StreamXout(self.node.id, self.stream)
        else:  # ColorCamera
            self.node.setVideoNumFramesPool(10)
            return StreamXout(self.node.id, self.stream)

    """
    Available outputs (to the host) of this component
    """

    class Out:
        _comp: 'CameraComponent'

        def __init__(self, cameraComponent: 'CameraComponent'):
            self._comp = cameraComponent

        def main(self, pipeline: dai.Pipeline, device: dai.Device) -> XoutBase:
            """
            Default output. Uses either camera(), replay(), or encoded() depending on the component settings.
            """
            if self._comp.encoder:
                return self.encoded(pipeline, device)
            elif self._comp.is_replay():
                return self.replay(pipeline, device)
            else:
                return self.camera(pipeline, device)

        def camera(self, pipeline: dai.Pipeline, device: dai.Device) -> XoutFrames:
            """
            Streams camera output to the OAK camera. Produces FramePacket.
            """
            out = XoutFrames(self._comp.get_stream_xout(), self._comp._get_fps())
            out.name = self._comp._source
            return self._comp._create_xout(pipeline, out)

        def replay(self, pipeline: dai.Pipeline, device: dai.Device) -> XoutBase:
            """
            If depthai-recording was used, it won't stream anything, but it will instead use frames that were sent to the OAK. Produces FramePacket.
            """
            out = XoutFrames(ReplayStream(self._comp._source), self._comp._get_fps())
            return self._comp._create_xout(pipeline, out)

        def encoded(self, pipeline: dai.Pipeline, device: dai.Device) -> XoutBase:
            """
            If encoding was enabled, it will stream bitstream from VideoEncoder node to the host. Produces FramePacket.
            """
            if self._comp._encoderProfile == dai.VideoEncoderProperties.Profile.MJPEG:
                out = XoutMjpeg(
                    frames=StreamXout(self._comp.encoder.id, self._comp.encoder.bitstream),
                    color=self._comp.is_color(),
                    lossless=self._comp.encoder.getLossless(),
                    fps=self._comp.encoder.getFrameRate(),
                    frame_shape=self._comp.stream_size
                )
            else:
                out = XoutH26x(
                    frames=StreamXout(self._comp.encoder.id, self._comp.encoder.bitstream),
                    color=self._comp.is_color(),
                    profile=self._comp._encoderProfile,
                    fps=self._comp.encoder.getFrameRate(),
                    frame_shape=self._comp.stream_size
                )
            out.name = self._comp._source
            return self._comp._create_xout(pipeline, out)

    out: Out



#!/usr/bin/python

def build_context_menu(graph):
    """
    populate the node graph's context menu with essential menus commands.

    Args:
        graph (NodeGraphQt.NodeGraph): node graph controller.
    """
    from Qt import QtGui, QtCore
    graph_menu = graph.get_context_menu('graph')

    # "File" menu.
    # --------------------------------------------------------------------------
    file_menu = graph_menu.add_menu('&File')

    file_menu.add_command('Open...', _open_session, QtGui.QKeySequence.Open)
    file_menu.add_command('Import...', _import_session)
    file_menu.add_command('Save...', _save_session, QtGui.QKeySequence.Save)
    file_menu.add_command('Save As...', _save_session_as, 'Ctrl+Shift+S')
    file_menu.add_command('New Session', _new_session)

    # "Edit" menu.
    # --------------------------------------------------------------------------
    edit_menu = graph_menu.add_menu('&Edit')
    # edit_menu.add_command("Add Backdrop Node", _add_backdrop_node, "B") # GX

    # edit_menu.add_separator()
    edit_menu.add_command('Clear Undo History', _clear_undo)
    edit_menu.add_command('Show Undo History', _show_undo_view)
    edit_menu.add_separator()

    # edit_menu.add_command('Copy', _copy_nodes, QtGui.QKeySequence.Copy)
    # edit_menu.add_command('Cut', _cut_nodes, QtGui.QKeySequence.Cut)
    # edit_menu.add_command('Paste', _paste_nodes, QtGui.QKeySequence.Paste)
    # edit_menu.add_command('Delete', _delete_nodes, QtGui.QKeySequence.Delete)

    # edit_menu.add_separator()

    edit_menu.add_command('Select all', _select_all_nodes, 'Ctrl+A')
    edit_menu.add_command('Deselect all', _clear_node_selection, 'Ctrl+Shift+A')
    # edit_menu.add_command('Enable/Disable', _disable_nodes, 'D')

    # edit_menu.add_command('Duplicate', _duplicate_nodes, 'Alt+C')
    edit_menu.add_command('Center Selection', _fit_to_selection, 'F')

    edit_menu.add_separator()

    edit_menu.add_command('Zoom In', _zoom_in, '=')
    edit_menu.add_command('Zoom Out', _zoom_out, '-')
    # edit_menu.add_command('Reset Zoom', _reset_zoom, 'H')

    edit_menu.add_separator()

    # "Grid Mode" submenu.
    # --------------------------------------------------------------------------
    bg_menu = edit_menu.add_menu('&Grid Mode')
    bg_menu.add_command('None', _bg_grid_none)
    bg_menu.add_command('Lines', _bg_grid_lines)
    bg_menu.add_command('Dots', _bg_grid_dots)

    edit_menu.add_separator()

    # "Node" menu.
    # --------------------------------------------------------------------------
    # node_menu = graph_menu.add_menu('&Nodes')
    # node_menu.add_command('Node Search', _toggle_node_search, 'Tab')
    # node_menu.add_separator()
    # node_menu.add_command(
    #     'Auto Layout Up Stream', _layout_graph_up, 'L')
    # node_menu.add_command(
    #     'Auto Layout Down Stream', _layout_graph_down, 'Ctrl+L')
    # node_menu.add_separator()
    # node_menu.add_command(
    #     'Expand Group Node', _expand_group_node,
    #     QtGui.QKeySequence(QtCore.Qt.ALT + QtCore.Qt.Key_Return))

    # "Pipe" menu.
    # --------------------------------------------------------------------------
    pipe_menu = graph_menu.add_menu('&Pipes')
    pipe_menu.add_command('Curved', _curved_pipe)
    pipe_menu.add_command('Straight', _straight_pipe)
    pipe_menu.add_command('Angle', _angle_pipe)

# ------------------------------------------------------------------------------
# --- menu command functions ---
# ------------------------------------------------------------------------------


def _zoom_in(graph):
    """
    Set the node graph to zoom in by 0.1
    """
    zoom = graph.get_zoom() + 0.1
    graph.set_zoom(zoom)


def _zoom_out(graph):
    """
    Set the node graph to zoom in by 0.1
    """
    zoom = graph.get_zoom() - 0.2
    graph.set_zoom(zoom)


def _reset_zoom(graph):
    """
    Reset zoom level.
    """
    graph.reset_zoom()


def _open_session(graph):
    """
    Prompts a file open dialog to load a session.
    """
    current = graph.current_session()
    file_path = graph.load_dialog(current)
    if file_path:
        graph.load_session(file_path)


def _import_session(graph):
    """
    Prompts a file open dialog to load a session.
    """
    current = graph.current_session()
    file_path = graph.load_dialog(current)
    if file_path:
        graph.import_session(file_path)


def _save_session(graph):
    """
    Prompts a file save dialog to serialize a session if required.
    """
    current = graph.current_session()
    if current:
        graph.save_session(current)
        msg = 'Session layout saved:\n{}'.format(current)
        viewer = graph.viewer()
        viewer.message_dialog(msg, title='Session Saved')
    else:
        _save_session_as(graph)


def _save_session_as(graph):
    """
    Prompts a file save dialog to serialize a session.
    """
    current = graph.current_session()
    file_path = graph.save_dialog(current)
    if file_path:
        graph.save_session(file_path)


def _new_session(graph):
    """
    Prompts a warning dialog to new a node graph session.
    """
    if graph.question_dialog('Clear Current Session?', 'Clear Session'):
        graph.clear_session()

def _add_backdrop_node(graph):
    """
    GX
    Create a backdrop node
    """
    viewer = graph._viewer
    pos = viewer.mapToScene(viewer._previous_pos)
    pos = (pos.x(), pos.y())
    graph.create_node("Backdrop", pos=pos)

def _clear_undo(graph):
    """
    Prompts a warning dialog to clear undo.
    """
    viewer = graph.viewer()
    msg = 'Clear all undo history, Are you sure?'
    if viewer.question_dialog('Clear Undo History', msg):
        graph.clear_undo_stack()


def _copy_nodes(graph):
    """
    Copy nodes to the clipboard.
    """
    graph.copy_nodes()


def _cut_nodes(graph):
    """
    Cut nodes to the clip board.
    """
    graph.cut_nodes()


def _paste_nodes(graph):
    """
    Pastes nodes copied from the clipboard.
    """
    graph.paste_nodes()


def _delete_nodes(graph):
    """
    Delete selected node.
    """
    graph.delete_nodes(graph.selected_nodes())


def _select_all_nodes(graph):
    """
    Select all nodes.
    """
    graph.select_all()


def _clear_node_selection(graph):
    """
    Clear node selection.
    """
    graph.clear_selection()


def _disable_nodes(graph):
    """
    Toggle disable on selected nodes.
    """
    graph.disable_nodes(graph.selected_nodes())


def _duplicate_nodes(graph):
    """
    Duplicated selected nodes.
    """
    graph.duplicate_nodes(graph.selected_nodes())


def _expand_group_node(graph):
    """
    Expand selected group node.
    """
    selected_nodes = graph.selected_nodes()
    if not selected_nodes:
        graph.message_dialog('Please select a "GroupNode" to expand.')
        return
    graph.expand_group_node(selected_nodes[0])


def _fit_to_selection(graph):
    """
    Sets the zoom level to fit selected nodes.
    """
    graph.fit_to_selection()


def _show_undo_view(graph):
    """
    Show the undo list widget.
    """
    graph.undo_view.show()


def _curved_pipe(graph):
    """
    Set node graph pipes layout as curved.
    """
    from NodeGraphQt.constants import PipeLayoutEnum
    graph.set_pipe_style(PipeLayoutEnum.CURVED.value)


def _straight_pipe(graph):
    """
    Set node graph pipes layout as straight.
    """
    from NodeGraphQt.constants import PipeLayoutEnum
    graph.set_pipe_style(PipeLayoutEnum.STRAIGHT.value)


def _angle_pipe(graph):
    """
    Set node graph pipes layout as angled.
    """
    from NodeGraphQt.constants import PipeLayoutEnum
    graph.set_pipe_style(PipeLayoutEnum.ANGLE.value)


def _bg_grid_none(graph):
    """
    Turn off the background patterns.
    """
    from NodeGraphQt.constants import ViewerEnum
    graph.set_grid_mode(ViewerEnum.GRID_DISPLAY_NONE.value)


def _bg_grid_dots(graph):
    """
    Set background node graph background with grid dots.
    """
    from NodeGraphQt.constants import ViewerEnum
    graph.set_grid_mode(ViewerEnum.GRID_DISPLAY_DOTS.value)


def _bg_grid_lines(graph):
    """
    Set background node graph background with grid lines.
    """
    from NodeGraphQt.constants import ViewerEnum
    graph.set_grid_mode(ViewerEnum.GRID_DISPLAY_LINES.value)


def _layout_graph_down(graph):
    """
    Auto layout the nodes down stream.
    """
    nodes = graph.selected_nodes() or graph.all_nodes()
    graph.auto_layout_nodes(nodes=nodes, down_stream=True)


def _layout_graph_up(graph):
    """
    Auto layout the nodes up stream.
    """
    nodes = graph.selected_nodes() or graph.all_nodes()
    graph.auto_layout_nodes(nodes=nodes, down_stream=False)


def _toggle_node_search(graph):
    """
    show/hide the node search widget.
    """
    graph.toggle_node_search()



import collections
import time

import cv2


class FPSHandler:
    """
    Class that handles all FPS-related operations. Mostly used to calculate different streams FPS, but can also be
    used to feed the video file based on it's FPS property, not app performance (this prevents the video from being sent
    to quickly if we finish processing a frame earlier than the next video frame should be consumed)
    """

    _fpsBgColor = (0, 0, 0)
    _fpsColor = (255, 255, 255)
    _fpsType = cv2.FONT_HERSHEY_SIMPLEX
    _fpsLineType = cv2.LINE_AA

    def __init__(self, cap=None, maxTicks=100):
        """
        Args:
            cap (cv2.VideoCapture, Optional): handler to the video file object
            maxTicks (int, Optional): maximum ticks amount for FPS calculation
        """
        self._timestamp = None
        self._start = None
        self._framerate = cap.get(cv2.CAP_PROP_FPS) if cap is not None else None
        self._useCamera = cap is None

        self._iterCnt = 0
        self._ticks = {}

        if maxTicks < 2:
            raise ValueError(f"Proviced maxTicks value must be 2 or higher (supplied: {maxTicks})")

        self._maxTicks = maxTicks

    def nextIter(self):
        """
        Marks the next iteration of the processing loop. Will use :obj:`time.sleep` method if initialized with video file
        object
        """
        if self._start is None:
            self._start = time.monotonic()

        if not self._useCamera and self._timestamp is not None:
            frameDelay = 1.0 / self._framerate
            delay = (self._timestamp + frameDelay) - time.monotonic()
            if delay > 0:
                time.sleep(delay)
        self._timestamp = time.monotonic()
        self._iterCnt += 1

    def tick(self, name):
        """
        Marks a point in time for specified name

        Args:
            name (str): Specifies timestamp name
        """
        if name not in self._ticks:
            self._ticks[name] = collections.deque(maxlen=self._maxTicks)
        self._ticks[name].append(time.monotonic())

    def tickFps(self, name):
        """
        Calculates the FPS based on specified name

        Args:
            name (str): Specifies timestamps' name

        Returns:
            float: Calculated FPS or :code:`0.0` (default in case of failure)
        """
        if name in self._ticks and len(self._ticks[name]) > 1:
            timeDiff = self._ticks[name][-1] - self._ticks[name][0]
            return (len(self._ticks[name]) - 1) / timeDiff if timeDiff != 0 else 0.0
        else:
            return 0.0

    def fps(self):
        """
        Calculates FPS value based on :func:`nextIter` calls, being the FPS of processing loop

        Returns:
            float: Calculated FPS or :code:`0.0` (default in case of failure)
        """
        if self._start is None or self._timestamp is None:
            return 0.0
        timeDiff = self._timestamp - self._start
        return self._iterCnt / timeDiff if timeDiff != 0 else 0.0

    def printStatus(self):
        """
        Prints total FPS for all names stored in :func:`tick` calls
        """
        print("=== TOTAL FPS ===")
        for name in self._ticks:
            print(f"[{name}]: {self.tickFps(name):.1f}")

    def drawFps(self, frame, name):
        """
        Draws FPS values on requested frame, calculated based on specified name

        Args:
            frame (numpy.ndarray): Frame object to draw values on
            name (str): Specifies timestamps' name
        """
        frameFps = f"{name.upper()} FPS: {round(self.tickFps(name), 1)}"
        # cv2.rectangle(frame, (0, 0), (120, 35), (255, 255, 255), cv2.FILLED)
        cv2.putText(frame, frameFps, (5, 15), self._fpsType, 0.5, self._fpsBgColor, 4, self._fpsLineType)
        cv2.putText(frame, frameFps, (5, 15), self._fpsType, 0.5, self._fpsColor, 1, self._fpsLineType)

        if "nn" in self._ticks:
            cv2.putText(frame, f"NN FPS:  {round(self.tickFps('nn'), 1)}", (5, 30), self._fpsType, 0.5,
                        self._fpsBgColor, 4, self._fpsLineType)
            cv2.putText(frame, f"NN FPS:  {round(self.tickFps('nn'), 1)}", (5, 30), self._fpsType, 0.5, self._fpsColor,
                        1, self._fpsLineType)



#!/usr/bin/python
from ..base.commands import PropertyChangedCmd
from ..base.model import NodeModel
from ..constants import (NODE_PROP,
                                   NODE_LAYOUT_DIRECTION,
                                   NODE_LAYOUT_VERTICAL,
                                   NODE_LAYOUT_HORIZONTAL)


class _ClassProperty(object):

    def __init__(self, f):
        self.f = f

    def __get__(self, instance, owner):
        return self.f(owner)


class NodeObject(object):
    """
    The ``NodeGraphQt.NodeObject`` class is the main base class that all
    nodes inherit from.

    **Inherited by:**
        :class:`NodeGraphQt.BaseNode`,
        :class:`NodeGraphQt.BackdropNode`

    Args:
        qgraphics_views (dict): Dictionary with the node layout type as the key
            and a custom graphics item subclassed from the ``AbstractNodeItem``
            as the value.

            .. code-block:: python

                # snippet taken from the NodeGraphQt.BaseNode class.

                class BaseNode(NodeObject):

                    def __init__(self, qgraphics_views=None):
                        qgraphics_views = qgraphics_views or {
                            NodeGraphQt.constants.NODE_LAYOUT_HORIZONTAL: NodeItem,
                            NodeGraphQt.constants.NODE_LAYOUT_VERTICAL: NodeItemVertical
                        }
                        super(BaseNode, self).__init__(qgraphics_views)

    """

    # Unique node identifier domain. `eg.` ``"com.chantacticvfx"``
    __identifier__ = 'nodeGraphQt.nodes'

    # Base node name.
    NODE_NAME = None

    def __init__(self, qgraphics_views=None):
        self._graph = None
        self._model = NodeModel()
        self._model.type_ = self.type_
        self._model.name = self.NODE_NAME

        _NodeItem = None
        if NODE_LAYOUT_DIRECTION is NODE_LAYOUT_VERTICAL:
            _NodeItem = qgraphics_views.get(NODE_LAYOUT_VERTICAL)
        elif NODE_LAYOUT_DIRECTION is NODE_LAYOUT_HORIZONTAL:
            _NodeItem = qgraphics_views.get(NODE_LAYOUT_HORIZONTAL)

        if _NodeItem is None:
            raise ValueError(
                'qgraphics item for the {} node layout can\'t be None!'.format({
                    NODE_LAYOUT_VERTICAL: 'vertical',
                    NODE_LAYOUT_HORIZONTAL: 'horizontal'
                }[NODE_LAYOUT_DIRECTION]))

        self._view = _NodeItem()
        self._view.type_ = self.type_
        self._view.name = self.model.name
        self._view.id = self._model.id

    def __repr__(self):
        return '<{}("{}") object at {}>'.format(
            self.__class__.__name__, self.NODE_NAME, hex(id(self)))

    @_ClassProperty
    def type_(cls):
        """
        Node type identifier followed by the class name.
        `eg.` ``"com.chantacticvfx.NodeObject"``

        Returns:
            str: node type.
        """
        return cls.__identifier__ + '.' + cls.__name__

    @property
    def id(self):
        """
        The node unique id.

        Returns:
            str: unique id string.
        """
        return self.model.id

    @property
    def graph(self):
        """
        The parent node graph.

        Returns:
            NodeGraphQt.NodeGraph: node graph.
        """
        return self._graph

    @property
    def view(self):
        """
        Returns the :class:`QtWidgets.QGraphicsItem` used in the scene.

        Returns:
            NodeGraphQt.qgraphics.node_abstract.AbstractNodeItem: node item.
        """
        return self._view

    def set_view(self, item):
        """
        Set a new ``QGraphicsItem`` item to be used as the view.
        (the provided qgraphics item must be subclassed from the
        ``AbstractNodeItem`` object.)

        Args:
            item (NodeGraphQt.qgraphics.node_abstract.AbstractNodeItem): node item.
        """
        if self._view:
            old_view = self._view
            scene = self._view.scene()
            scene.removeItem(old_view)
            self._view = item
            scene.addItem(self._view)
        else:
            self._view = item
        self.NODE_NAME = self._view.name

        # update the view.
        self.update()

    @property
    def model(self):
        """
        Return the node model.

        Returns:
            NodeGraphQt.base.model.NodeModel: node model object.
        """
        return self._model

    def set_model(self, model):
        """
        Set a new model to the node model.
        (Setting a new node model will also update the views qgraphics item.)

        Args:
            model (NodeGraphQt.base.model.NodeModel): node model object.
        """
        self._model = model
        self._model.type_ = self.type_
        self._model.id = self.view.id

        # update the view.
        self.update()

    def update_model(self):
        """
        Update the node model from view.
        """
        for name, val in self.view.properties.items():
            if name in self.model.properties.keys():
                setattr(self.model, name, val)
            if name in self.model.custom_properties.keys():
                self.model.custom_properties[name] = val

    def update(self):
        """
        Update the node view from model.
        """
        settings = self.model.to_dict[self.model.id]
        settings['id'] = self.model.id
        if settings.get('custom'):
            settings['widgets'] = settings.pop('custom')

        self.view.from_dict(settings)

    def serialize(self):
        """
        Serialize node model to a dictionary.

        example:

        .. highlight:: python
        .. code-block:: python

            {'0x106cf75a8': {
                'name': 'foo node',
                'color': (48, 58, 69, 255),
                'border_color': (85, 100, 100, 255),
                'text_color': (255, 255, 255, 180),
                'type': 'com.chantasticvfx.MyNode',
                'selected': False,
                'disabled': False,
                'visible': True,
                'inputs': {
                    <port_name>: {<node_id>: [<port_name>, <port_name>]}
                },
                'outputs': {
                    <port_name>: {<node_id>: [<port_name>, <port_name>]}
                },
                'input_ports': [<port_name>, <port_name>],
                'output_ports': [<port_name>, <port_name>],
                'width': 0.0,
                'height: 0.0,
                'pos': (0.0, 0.0),
                'custom': {},
                }
            }

        Returns:
            dict: serialized node
        """
        return self.model.to_dict

    def name(self):
        """
        Name of the node.

        Returns:
            str: name of the node.
        """
        return self.model.name

    def set_name(self, name=''):
        """
        Set the name of the node.

        Args:
            name (str): name for the node.
        """
        self.set_property('name', name)

    def color(self):
        """
        Returns the node color in (red, green, blue) value.

        Returns:
            tuple: ``(r, g, b)`` from ``0-255`` range.
        """
        r, g, b, a = self.model.color
        return r, g, b

    def set_color(self, r=0, g=0, b=0):
        """
        Sets the color of the node in (red, green, blue) value.

        Args:
            r (int): red value ``0-255`` range.
            g (int): green value ``0-255`` range.
            b (int): blue value ``0-255`` range.
        """
        self.set_property('color', (r, g, b, 255))

    def disabled(self):
        """
        Returns whether the node is enabled or disabled.

        Returns:
            bool: True if the node is disabled.
        """
        return self.model.disabled

    def set_disabled(self, mode=False):
        """
        Set the node state to either disabled or enabled.

        Args:
            mode(bool): True to disable node.
        """
        self.set_property('disabled', mode)

    def selected(self):
        """
        Returns the selected state of the node.

        Returns:
            bool: True if the node is selected.
        """
        self.model.selected = self.view.isSelected()
        return self.model.selected

    def set_selected(self, selected=True):
        """
        Set the node to be selected or not selected.

        Args:
            selected (bool): True to select the node.
        """
        self.set_property('selected', selected)

    def create_property(self, name, value, items=None, range=None,
                        widget_type=NODE_PROP, tab=None):
        """
        Creates a custom property to the node.

        See Also:
            Custom node properties bin widget
            :class:`NodeGraphQt.PropertiesBinWidget`

        Hint:
            Here are some constants variables used to define the node
            widget type in the ``PropertiesBinWidget``.

            - :attr:`NodeGraphQt.constants.NODE_PROP`
            - :attr:`NodeGraphQt.constants.NODE_PROP_QLABEL`
            - :attr:`NodeGraphQt.constants.NODE_PROP_QLINEEDIT`
            - :attr:`NodeGraphQt.constants.NODE_PROP_QTEXTEDIT`
            - :attr:`NodeGraphQt.constants.NODE_PROP_QCOMBO`
            - :attr:`NodeGraphQt.constants.NODE_PROP_QCHECKBOX`
            - :attr:`NodeGraphQt.constants.NODE_PROP_QSPINBOX`
            - :attr:`NodeGraphQt.constants.NODE_PROP_COLORPICKER`
            - :attr:`NodeGraphQt.constants.NODE_PROP_FILE`
            - :attr:`NodeGraphQt.constants.NODE_PROP_VECTOR2`
            - :attr:`NodeGraphQt.constants.NODE_PROP_VECTOR3`
            - :attr:`NodeGraphQt.constants.NODE_PROP_VECTOR4`
            - :attr:`NodeGraphQt.constants.NODE_PROP_FLOAT`
            - :attr:`NodeGraphQt.constants.NODE_PROP_INT`
            - :attr:`NodeGraphQt.constants.NODE_PROP_BUTTON`

        Args:
            name (str): name of the property.
            value (object): data.
            items (list[str]): items used by widget type ``NODE_PROP_QCOMBO``
            range (tuple)): ``(min, max)`` values used by ``NODE_PROP_SLIDER``
            widget_type (int): widget flag to display in the
                :class:`NodeGraphQt.PropertiesBinWidget`
            tab (str): name of the widget tab to display in the properties bin.
        """
        self.model.add_property(name, value, items, range, widget_type, tab)

    def properties(self):
        """
        Returns all the node properties.

        Returns:
            dict: a dictionary of node properties.
        """
        props = self.model.to_dict[self.id].copy()
        props['id'] = self.id
        return props

    def get_property(self, name):
        """
        Return the node custom property.

        Args:
            name (str): name of the property.

        Returns:
            object: property data.
        """
        if self.graph and name == 'selected':
            self.model.set_property(name, self.view.selected)

        return self.model.get_property(name)

    def set_property(self, name, value, push_undo=True):
        """
        Set the value on the node custom property.

        Args:
            name (str): name of the property.
            value (object): property data (python built in types).
            push_undo (bool): register the command to the undo stack. (default: True)
        """

        # prevent signals from causing a infinite loop.
        if self.get_property(name) == value:
            return

        if self.graph and name == 'name':
            value = self.graph.get_unique_name(value)
            self.NODE_NAME = value

        if self.graph:
            if push_undo:
                undo_stack = self.graph.undo_stack()
                undo_stack.push(PropertyChangedCmd(self, name, value))
            else:
                PropertyChangedCmd(self, name, value).redo()
        else:
            if hasattr(self.view, name):
                setattr(self.view, name, value)
            self.model.set_property(name, value)

    def has_property(self, name):
        """
        Check if node custom property exists.

        Args:
            name (str): name of the node.

        Returns:
            bool: true if property name exists in the Node.
        """
        return name in self.model.custom_properties.keys()

    def set_x_pos(self, x):
        """
        Set the node horizontal X position in the node graph.

        Args:
            x (float or int): node X position.
        """
        y = self.pos()[1]
        self.set_pos(float(x), y)

    def set_y_pos(self, y):
        """
        Set the node horizontal Y position in the node graph.

        Args:
            y (float or int): node Y position.
        """

        x = self.pos()[0]
        self.set_pos(x, float(y))

    def set_pos(self, x, y):
        """
        Set the node X and Y position in the node graph.

        Args:
            x (float or int): node X position.
            y (float or int): node Y position.
        """
        self.set_property('pos', [float(x), float(y)])

    def x_pos(self):
        """
        Get the node X position in the node graph.

        Returns:
            float: x position.
        """
        return self.model.pos[0]

    def y_pos(self):
        """
        Get the node Y position in the node graph.

        Returns:
            float: y position.
        """
        return self.model.pos[1]

    def pos(self):
        """
        Get the node XY position in the node graph.

        Returns:
            list[float, float]: x, y position.
        """
        if self.view.xy_pos and self.view.xy_pos != self.model.pos:
            self.model.pos = self.view.xy_pos

        return self.model.pos



from abc import ABC, abstractmethod
from enum import IntEnum
from pathlib import Path
from typing import List

import depthai as dai

import depthai_sdk.oak_outputs.xout as outputs
from depthai_sdk.oak_outputs.xout_base import XoutBase


class Recorder(ABC):
    @abstractmethod
    def write(self, name: str, frame: dai.ImgFrame):
        raise NotImplementedError()

    @abstractmethod
    def close(self):
        raise NotImplementedError()

    @abstractmethod
    def update(self, path: Path, device: dai.Device, xouts: List['XoutFrames']):
        raise NotImplementedError()


class OakStream:
    class StreamType(IntEnum):
        RAW = 0  # Unencoded frames (mono, color, disparity)
        MJPEG = 1
        H264 = 2
        H265 = 3
        DEPTH = 4  # 16 bit
        IMU = 5

    type: StreamType
    xlink_name: str

    def __init__(self, xout: XoutBase):
        if isinstance(xout, outputs.XoutMjpeg):
            self.type = self.StreamType.MJPEG
            self.xlink_name = xout.frames.name
        elif isinstance(xout, outputs.XoutH26x):
            self.xlink_name = xout.frames.name
            if xout.profile == dai.VideoEncoderProperties.Profile.H265_MAIN:
                self.type = self.StreamType.H265
            else:
                self.type = self.StreamType.H264
        elif isinstance(xout, outputs.XoutDepth):
            self.xlink_name = xout.frames.name
            self.type = self.StreamType.DEPTH  # TODO is depth raw or should it be DEPTH?
        elif isinstance(xout, outputs.XoutDisparity):
            self.xlink_name = xout.frames.name
            self.type = self.StreamType.RAW
        elif isinstance(xout, outputs.XoutFrames):
            self.xlink_name = xout.frames.name
            self.type = self.StreamType.RAW
        elif isinstance(xout, outputs.XoutIMU):
            self.xlink_name = xout.imu_out.name
            self.type = self.StreamType.IMU
        else:
            raise ValueError("You have passed invalid Component Output to the Recorder!")

    def fourcc(self) -> str:
        if self.type == self.StreamType.MJPEG:
            return 'mjpeg'
        elif self.type == self.StreamType.H264:
            return 'h264'
        elif self.type == self.StreamType.H265:
            return 'hevc'
        elif self.type == self.StreamType.DEPTH:
            return 'y16'

    def isH265(self) -> bool:
        return self.type == self.StreamType.H265

    def isH264(self) -> bool:
        return self.type == self.StreamType.H264

    def isH26x(self) -> bool:
        return self.isH264() or self.isH265()

    def isMjpeg(self) -> bool:
        return self.type == self.StreamType.MJPEG

    def isRaw(self) -> bool:
        return self.type == self.StreamType.RAW

    def isDepth(self) -> bool:
        return self.type == self.StreamType.DEPTH

    def isIMU(self):
        return self.type == self.StreamType.IMU



#!/usr/bin/python
# -*- coding: utf-8 -*-
__version__ = '0.2.2'
__status__ = 'Work in Progress'
__license__ = 'MIT'

__author__ = 'Johnny Chan'

__module_name__ = 'NodeGraphQt'
__url__ = 'https://github.com/jchanvfx/NodeGraphQt'



from pathlib import Path


def create_writer_dir(path: Path, name: str, extension: str = 'avi') -> str:
    if path.suffix == '':  # If path is a folder
        path.mkdir(parents=True, exist_ok=True)
        return str(path / f'{name}.{extension}')
    else:  # If path is a file
        path.parent.mkdir(parents=True, exist_ok=True)
        return str(path)



from .component import Component
from .camera_component import CameraComponent
from .nn_component import NNComponent
from .stereo_component import StereoComponent
from .imu_component import IMUComponent



#!/usr/bin/env python3
from argparse import ArgumentParser
import depthai as dai
import subprocess
import os, signal
import re
from Qt import QtWidgets, QtCore
from .NodeGraphQt import NodeGraph, BaseNode, PropertiesBinWidget
from .NodeGraphQt.constants import ViewerEnum
import json
import time
from threading import Thread
from typing import Any, Dict, List


class DepthaiNode(BaseNode):
    # unique node identifier.
    __identifier__ = 'dai'

    # initial default node name.
    NODE_NAME = 'Node'

    def __init__(self):
        super(DepthaiNode, self).__init__()
        # create QLineEdit text input widget.
        # self.add_text_input('my_input', 'Text Input', tab='widgets')

class TraceEvent():
    event: int
    status: int
    src_id: str
    dst_id: str
    timestamp = 0.0
    host_timestamp = 0.0

class NodePort:
    id: str # Id of the port
    name: str # preview, out, video...
    port: Any = None  # QT port object
    node: Dict  # From json schema
    type: int  # Input or output
    dai_node: Any
    group_name: str
    blocking: bool
    queue_size: int

    def nice_name(self) -> str: # For visualization
        return f"{self.group_name}[{self.name}]" if self.group_name else self.name

    def create(self) -> bool:
        return self.port is None
    def is_input(self) -> bool:
        return self.type == 3

    def is_output(self) -> bool:
        return self.type == 0

    def find_node(self, node_id: int, group_name: str, port_name: str) -> bool:
        return self.name == port_name and self.dai_node['id'] == node_id and self.group_name == group_name

    def __str__(self):
        return f"{self.dai_node['name']}.{self.name} ({self.id})"


class PipelineGraph:

    node_color = {
        "ColorCamera": (241, 148, 138),
        "MonoCamera": (243, 243, 243),
        "ImageManip": (174, 214, 241),
        "VideoEncoder": (190, 190, 190),

        "NeuralNetwork": (171, 235, 198),
        "DetectionNetwork": (171, 235, 198),
        "MobileNetDetectionNetwork": (171, 235, 198),
        "MobileNetSpatialDetectionNetwork": (171, 235, 198),
        "YoloDetectionNetwork": (171, 235, 198),
        "YoloSpatialDetectionNetwork": (171, 235, 198),
        "SpatialDetectionNetwork": (171, 235, 198),

        "SPIIn": (242, 215, 213),
        "XLinkIn": (242, 215, 213),

        "SPIOut": (230, 176, 170),
        "XLinkOut": (230, 176, 170),

        "Script": (249, 231, 159),

        "StereoDepth": (215, 189, 226),
        "SpatialLocationCalculator": (215, 189, 226),

        "EdgeDetector": (248, 196, 113),
        "FeatureTracker": (248, 196, 113),
        "ObjectTracker": (248, 196, 113),
        "IMU": (248, 196, 113)
    }
    default_node_color = (190, 190, 190)  # For node types that does not appear in 'node_color'
    process = None
    links: Dict[str, Dict[str, Any]]

    def __init__(self):
        # handle SIGINT to make the app terminate on CTRL+C
        signal.signal(signal.SIGINT, signal.SIG_DFL)

        QtCore.QCoreApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling)

        self.app = QtWidgets.QApplication(["DepthAI Pipeline Graph"])

        # create node graph controller.
        self.graph = NodeGraph()
        self.graph.set_background_color(255,255,255)
        self.graph.set_grid_mode(ViewerEnum.GRID_DISPLAY_NONE.value)

        self.graph.register_node(DepthaiNode)

        # TODO: show node properties via properties bin widget
        self.properties_bin = PropertiesBinWidget(node_graph=self.graph)
        self.properties_bin.setWindowFlags(QtCore.Qt.Tool)

        # show the node properties bin widget when a node is double-clicked.
        def display_properties_bin(node):
            if not self.properties_bin.isVisible():
                for p in self.ports:
                    if p.node == node:
                        txt = json.dumps(p.dai_node, sort_keys=True, indent=4, separators=(',', ': '))
                        self.properties_bin.node_clicked(txt)
                        self.properties_bin.show()
                        return

        # wire function to "node_double_clicked" signal.
        self.graph.node_double_clicked.connect(display_properties_bin)

        # show the node graph widget.
        self.graph_widget = self.graph.widget
        self.graph_widget.resize(1100, 800)

    def cmd_tool(self, args):
        if args.action == "load":
            self.graph_widget.show()
            self.graph.load_session(args.json)
            self.graph.fit_to_selection()
            self.graph.set_zoom(-0.9)
            self.graph.clear_selection()
            self.graph.clear_undo_stack()
            self.app.exec_()

        else:
            if args.action == "run":
                os.environ["PYTHONUNBUFFERED"] = "1"
                os.environ["DEPTHAI_LEVEL"] = "trace"

                command = args.command.split()
                if args.use_variable_names:
                    # If command starts with "python", we remove it
                    if "python" in command[0]:
                        command.pop(0)

                    command = "python -m trace -t ".split() + command

                pipeline_create_re = f'.*:\s*(.*)\s*=\s*{args.pipeline_name}\.create.*'
                node_name = []
                self.process = subprocess.Popen(command, shell=False, text=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

                schema_str = None
                record_output = "" # Save the output and print it in case something went wrong
                while True:
                    if self.process.poll() is not None: break
                    line = self.process.stdout.readline()
                    record_output += line
                    if args.verbose:
                        print(line.rstrip('\n'))
                    # we are looking for  a line:  ... [debug] Schema dump: {"connections":[{"node1Id":1...
                    match = re.match(r'.* Schema dump: (.*)', line)
                    if match:
                        schema_str = match.group(1)
                        print("Pipeline schema retrieved")
                        break
                        # TODO(themarpe) - expose "do kill" now
                        # # print(schema_str)
                        # if not args.do_not_kill:
                        #     print("Terminating program...")
                        #     process.terminate()
                    elif args.use_variable_names:
                        match = re.match(pipeline_create_re, line)
                        if match:
                            node_name.append(match.group(1))
                print("Program exited.")
                if schema_str is None:
                    if not args.verbose:
                        print(record_output)
                    print("\nSomething went wrong, the schema could not be extracted")
                    exit(1)
                schema = json.loads(schema_str)

            elif args.action == "from_file":
                with open(args.schema_file, "r") as schema_file:
                    # schema_file is either:
                    # 1) a Json file generated by a call to pipeline.serializeToJson(),
                    # 2) a log file generated by running the user program with DEPTHAI_LEVEL set to debug

                    # Are we in case 1) ?
                    try:
                        schema = json.load(schema_file)
                        if 'pipeline' not in schema:
                            print(f"Json file '{args.schema_file}' is missing 'pipeline' key")
                            exit(1)
                        schema = schema['pipeline']
                        print("Pipeline schema retrieved")
                    except json.decoder.JSONDecodeError:
                        # schema_file is not a Json file, so we are probably in case 2)
                        # we are looking for  a line:  ... [debug] Schema dump: {"connections":[{"node1Id":1...
                        schema_file.seek(0) # Rewind the file
                        while True:
                            line = schema_file.readline()
                            if len(line) == 0:
                                # End of file
                                print("\nSomething went wrong, the schema could not be extracted")
                                exit(1)
                            match = re.match(r'.* Schema dump: (.*)', line)
                            if match:
                                schema_str = match.group(1)
                                print("Pipeline schema retrieved")
                                break
                        schema = json.loads(schema_str)

            if args.verbose:
                print('Schema:', schema)

            self.create_graph(schema)

    def new_trace_log(self, msg: dai.LogMessage):
        self.new_trace_text(msg.payload)
    def new_trace_text(self, txt):
        # we are looking for  a line: EV:  ...
        match = re.search(r'EV:([0-9]+),S:([0-9]+),IDS:([0-9]+),IDD:([0-9]+),TSS:([0-9]+),TSN:([0-9]+)', txt.rstrip('\n'))
        if match:
            trace_event = TraceEvent()

            trace_event.event = int(match.group(1)) # 0=SEND, 1=RECEIVE, #2=PULL
            trace_event.status = int(match.group(2)) # 0=START, 1=END, 2=TIMEOUT
            trace_event.src_id = match.group(3)
            trace_event.dst_id = match.group(4)
            trace_event.timestamp = int(match.group(5)) + (int(match.group(6)) / 1000000000.0)
            trace_event.host_timestamp = time.time()

            if trace_event.status == 0 and trace_event.event == 0: # START event, send
                link = self.links[trace_event.dst_id][trace_event.src_id]
                link.new_event(trace_event.event)
            elif trace_event.status == 1 and trace_event.event == 1: # END event, receive
                for id, link in self.links[trace_event.src_id].items():
                    # SrcId is none, we can just take the first dst it
                    link.new_event(trace_event.event)
                    break
            elif trace_event.status == 2:
                print("TIMEOUT", txt)


    def traceEventReader(self):
        # local_event_buffer = []
        while self.process.poll() is None:
            line = self.process.stdout.readline()
            self.new_trace_text(line)

    def create_graph(self, schema: Dict, device: dai.Device = None):

        dai_connections = schema['connections']

        self.ports: List[NodePort] = []
        start_nodes = []
        for n in schema['nodes']:
            dict_n = n[1]
            dict_n['ioInfo'] = [el[1] for el in dict_n['ioInfo']]

            node_name = dict_n['name']

            # Create the node
            qt_node = self.graph.create_node('dai.DepthaiNode',
                                             name=node_name,
                                             selected=False,
                                             color=self.node_color.get(node_name, self.default_node_color),
                                             text_color=(0,0,0),
                                             push_undo=False)

            if node_name in ['ColorCamera', 'MonoCamera', 'XLinkIn']:
                start_nodes.append(qt_node)

            # Alphabetic order
            ioInfo = list(sorted(dict_n['ioInfo'], key = lambda el: el['name']))

            for dict_io in ioInfo:
                p = NodePort()
                p.name = dict_io['name']
                p.type = dict_io['type'] # Input/Output
                p.node = qt_node
                p.dai_node = n[1]
                p.id = str(dict_io['id'])
                p.group_name = dict_io['group']
                p.blocking = dict_io['blocking']
                p.queue_size = dict_io['queueSize']
                self.ports.append(p)

        self.links = dict()
        for i, c in enumerate(dai_connections):
            src_node_id = c["node1Id"]
            src_name = c["node1Output"]
            src_group = c["node1OutputGroup"]
            dst_node_id = c["node2Id"]
            dst_name = c["node2Input"]
            dst_group = c["node2InputGroup"]

            src_port = [p for p in self.ports if p.find_node(src_node_id, src_group, src_name)][0]
            dst_port = [p for p in self.ports if p.find_node(dst_node_id, dst_group, dst_name)][0]

            if src_port.create():  # Output
                src_port.port = src_port.node.add_output(name=src_port.nice_name(), color=(50,50,255))
            if dst_port.create(): # Input
                port_color = (249, 75, 0) if dst_port.blocking else (0, 255, 0)
                port_label = f"[{dst_port.queue_size}] {dst_port.nice_name()}"
                dst_port.port = dst_port.node.add_input(name=port_label, color=port_color, multi_input=True)

            print(f"{i}. {src_port} -> {dst_port}")
            link = src_port.port.connect_to(dst_port.port, push_undo=False)

            if dst_port.id not in self.links:
                self.links[dst_port.id] = {}
            self.links[dst_port.id][src_port.id] = link

        # Lock the ports
        self.graph.lock_all_ports()

        self.graph_widget.show()
        self.graph.auto_layout_nodes(start_nodes=start_nodes)
        self.graph.fit_to_selection()
        self.graph.set_zoom(-0.9)
        self.graph.clear_selection()
        self.graph.clear_undo_stack()

        self.app.processEvents()
        self.graph.auto_layout_nodes()


        if self.process is not None: # Arg tool
            reading_thread = Thread(target=self.traceEventReader, args=())
            reading_thread.start()
        else:
            # device.setLogOutputLevel(dai.LogLevel.TRACE)
            device.setLogLevel(dai.LogLevel.TRACE)
            device.addLogCallback(self.new_trace_log)

    def update(self): # Called by main loop (on main Thread)
        for _, ports in self.links.items():
            for _, link in ports.items():
                link.update_fps() # Update text

        self.app.processEvents()

def main():
    parser = ArgumentParser()
    subparsers = parser.add_subparsers(help="Action", required=True, dest="action")

    run_parser = subparsers.add_parser("run",
                                       help="Run your depthai program to create the corresponding pipeline graph")
    run_parser.add_argument('command', type=str,
                            help="The command with its arguments between ' or \" (ex: python script.py -i file)")
    run_parser.add_argument("-dnk", "--do_not_kill", action="store_true",
                            help="Don't terminate the command when the schema string has been retrieved")
    run_parser.add_argument("-var", "--use_variable_names", action="store_true",
                            help="Use the variable names from the python code to name the graph nodes")
    run_parser.add_argument("-p", "--pipeline_name", type=str, default="pipeline",
                            help="Name of the pipeline variable in the python code (default=%(default)s)")
    run_parser.add_argument('-v', '--verbose', action="store_true",
                            help="Show on the console the command output")

    from_file_parser = subparsers.add_parser("from_file",
                                             help="Create the pipeline graph by parsing a file containing the schema (log file generated with DEPTHAI_LEVEL=debug or Json file generated by pipeline.serializeToJSon())")
    from_file_parser.add_argument("schema_file",
                                  help="Path of the file containing the schema")

    load_parser = subparsers.add_parser("load", help="Load a previously saved pipeline graph")
    load_parser.add_argument("json_file",
                             help="Path of the .json file")
    args = parser.parse_args()

    p = PipelineGraph()
    p.cmd_tool(args)

    while True:
        p.update()
        time.sleep(0.001)

# Run as standalone tool
if __name__ == "__main__":
    main()



import math
import depthai as dai
from typing import *


class ImageSensor:
    name: str
    resolutions: List[Union[
        dai.ColorCameraProperties.SensorResolution,
        dai.MonoCameraProperties.SensorResolution
    ]]
    type: dai.node

    def __eq__(self, other):
        return self.name == other

    def __init__(self,
                 name: str,
                 resolutions: List[str],
                 type: str):
        from .parser import parse_resolution
        self.name = name
        self.type = dai.node.ColorCamera if type == 'color' else dai.node.MonoCamera
        self.resolutions = [parse_resolution(self.type, resolution) for resolution in resolutions]

    @property
    def maxRes(self) -> Union[dai.ColorCameraProperties.SensorResolution, dai.MonoCameraProperties.SensorResolution]:
        return self.resolutions[0]


cameraSensors: List[ImageSensor] = [
    ImageSensor('IMX378', ['12mp', '4k', '1080p'], 'color'),
    ImageSensor('OV9282', ['800P', '720p', '400p'], 'mono'),
    ImageSensor('OV9782', ['800P', '720p', '400p'], 'color'),
    ImageSensor('OV9281', ['800P', '720p', '400p'], 'color'),
    ImageSensor('IMX214', ['13mp', '12mp', '4k', '1080p'], 'color'),
    ImageSensor('OV7750', ['480P', '400p'], 'mono'),
    ImageSensor('OV7251', ['480P', '400p'], 'mono'),
    ImageSensor('IMX477', ['12mp', '4k', '1080p'], 'color'),
    ImageSensor('IMX577', ['12mp', '4k', '1080p'], 'color'),
    ImageSensor('AR0234', ['1200P'], 'color'),
    ImageSensor('IMX582', ['48mp', '12mp', '4k'], 'color'),
]

cameraResolutions: Dict[Any, Tuple[int, int]] = {
    dai.ColorCameraProperties.SensorResolution.THE_13_MP: (4208, 3120),
    dai.ColorCameraProperties.SensorResolution.THE_12_MP: (4056, 3040),
    dai.ColorCameraProperties.SensorResolution.THE_4_K: (3840, 2160),
    dai.ColorCameraProperties.SensorResolution.THE_1080_P: (1920, 1080),
    dai.ColorCameraProperties.SensorResolution.THE_800_P: (1280, 800),
    dai.ColorCameraProperties.SensorResolution.THE_720_P: (1280, 720),

    dai.MonoCameraProperties.SensorResolution.THE_800_P: (1280, 800),
    dai.MonoCameraProperties.SensorResolution.THE_720_P: (1280, 720),
    dai.MonoCameraProperties.SensorResolution.THE_480_P: (640, 480),
    dai.MonoCameraProperties.SensorResolution.THE_400_P: (640, 400),
}


def availableIspScales() -> List[Tuple[int, Tuple[int, int]]]:
    """
    Calculates all supported
    @rtype: List[ratio, [Numerator, Denominator]]
    """
    lst = []
    for n in range(1, 16 + 1):
        for d in range(n, 63 + 1):
            # Chroma needs 2x extra downscaling
            if d < 32 or n % 2 == 0:
                # Only if irreducible
                if math.gcd(n, d) == 1:
                    lst.append((n / d, (n, d)))
    lst.sort(reverse=True)
    return lst


def getClosestVideoSize(width: int, height: int, videoEncoder: bool=False) -> Tuple[int, int]:
    """
    For colorCamera.video output
    """
    while True:
        if width % 3 == 0:
            if not videoEncoder or width % 32 == 0:
                break
        width -= 1
    while True:
        if height % 2 == 0:
            break
        height -= 1
    return (width, height)


def getClosestIspScale(camResolution: Tuple[int, int],
                       width: Optional[int] = None,
                       height: Optional[int] = None,
                       videoEncoder: bool = False,
                       encoderFlag: bool = True,
                       ) -> List[int]:
    """
    Provides the closest ISP scaling values to either specified width or height.
    @param camResolution: Resolution (W, H) of the ColorCamera
    @param width: Desired width after ISP scaling. Conflicts with height
    @param height: Desired height after ISP scaling. Conflicts with width
    @param videoEncoder: If we want to stream ISP output directly into VideoEncoder (take into account its limitations).
    Width and height scaling values won't be the same (ignore aspect ratio), as otherwise only 5 ISP scaling options are
    available.
    @param encoderFlag: Not for user. Flag to avoid infinite looping.
    @return ISP scaling values (list of 4 ints)
    """
    if width and height:
        raise ValueError("You have to specify EITHER width OR height to calculate desired ISP scaling options!")
    if not width and not height:
        raise ValueError("You have to provide width or height calculate desired ISP scaling options!")

    minError = 99999
    ispScale: List[int] = None
    for ratio, (n, d) in availableIspScales():
        newW = int((camResolution[0] * n - 1) / d + 1)
        newH = int((camResolution[1] * n - 1) / d + 1)

        if (videoEncoder and
                (width and newW % 32 != 0 or
                 height and newH % 8 != 0)):
            continue  # ISP output size isn't supported by VideoEncoder

        # Currently, new ISP width must be divisible by 2. FW engineers are looking into it.
        if newW % 2 != 0:
            continue

        err = abs((newW - width) if width else (newH - height))
        if err < minError:
            ispScale = [n, d, n, d]
            minError = err

    if videoEncoder and encoderFlag:
        # Calculate the ISP scale for the other direction. Note that this means aspect ratio won't be preserved
        if width:
            hScale = getClosestIspScale(camResolution,
                                        height=int(camResolution[1] * ispScale[0] / ispScale[1]),
                                        videoEncoder=True,
                                        encoderFlag=False  # To avoid infinite loop
                                        )
            ispScale[2] = hScale[2]
            ispScale[3] = hScale[3]
        else:
            wScale = getClosestIspScale(camResolution,
                                        width=int(camResolution[0] * ispScale[0] / ispScale[1]),
                                        videoEncoder=True,
                                        encoderFlag=False  # To avoid infinite loop
                                        )
            ispScale[0] = wScale[0]
            ispScale[1] = wScale[1]

    return ispScale


def setCameraControl(control: dai.CameraControl,
                     manualFocus: Optional[int] = None,
                     afMode: Optional[dai.CameraControl.AutoFocusMode] = None,
                     awbMode: Optional[dai.CameraControl.AutoWhiteBalanceMode] = None,
                     sceneMode: Optional[dai.CameraControl.SceneMode] = None,
                     antiBandingMode: Optional[dai.CameraControl.AntiBandingMode] = None,
                     effectMode: Optional[dai.CameraControl.EffectMode] = None,
                     sharpness: Optional[int] = None,
                     lumaDenoise: Optional[int] = None,
                     chromaDenoise: Optional[int] = None,
                     ):
    """
    This function will be used when initializing cameras (ColorCamera.initialControl) and during runtime,
    when user configures camera with keyboard (if enabled).
    """
    if manualFocus is not None:
        control.setManualFocus(manualFocus)
    if afMode is not None:
        control.setAutoFocusMode(afMode)
    if awbMode is not None:
        control.setAutoWhiteBalanceMode(awbMode)
    if sceneMode is not None:
        control.setSceneMode(sceneMode)
    if antiBandingMode is not None:
        control.setAntiBandingMode(antiBandingMode)
    if effectMode is not None:
        control.setEffectMode(effectMode)
    if sharpness is not None:
        control.setSharpness(sharpness)
    if lumaDenoise is not None:
        control.setLumaDenoise(lumaDenoise)
    if chromaDenoise is not None:
        control.setChromaDenoise(chromaDenoise)

    # TODO: Add contrast, exposure compensation, brightness, manual exposure, and saturation


def cameraSensor(sensorName: str) -> ImageSensor:
    return cameraSensors[cameraSensors.index(sensorName.upper())]


def cameraSensorType(sensorName: str) -> dai.node:
    """
    Gets camera sensor type from it's name, either MonoCamera or ColorCamera.
    @param sensorName: Name of the camera sensor
    @return: dai.node.MonoCamera or dai.node.ColorCamera
    """
    return cameraSensor(sensorName).type


def cameraSensorResolution(sensorName: str
                           ) -> Union[
    dai.ColorCameraProperties.SensorResolution, dai.MonoCameraProperties.SensorResolution]:
    """
    Gets camera sensor type from it's name, either MonoCamera or ColorCamera.
    @param sensorName: Name of the camera sensor
    @return: dai.node.MonoCamera or dai.node.ColorCamera
    """
    return cameraSensor(sensorName).maxRes


def cameraSensorResolutionSize(sensorName: str) -> Tuple[int, int]:
    res = cameraSensorResolution(sensorName)
    return cameraResolutions[res]


def getClosesResolution(sensorName: str,
                        width: Optional[int] = None,
                        height: Optional[int] = None, ):
    if width and height:
        raise ValueError("You have to specify EITHER width OR height to calculate desired ISP scaling options!")
    if not width and not height:
        raise ValueError("You have to provide width or height calculate desired ISP scaling options!")

    minError = 99999
    closestRes = None
    desired, i = (width, 0) if width else (height, 1)
    for res in cameraSensor(sensorName).resolutions:
        size = cameraResolutions[res]
        err = abs(size[i] - desired)
        if err < minError:
            minError = err
            closestRes = res
    return closestRes


def getResize(size: Tuple[int, int],
                    width: Optional[int] = None,
                    height: Optional[int] = None) -> Tuple[int, int]:
    if width and height:
        raise ValueError("You have to specify EITHER width OR height to calculate desired ISP scaling options!")
    if not width and not height:
        raise ValueError("You have to provide width or height calculate desired ISP scaling options!")

    if width:
        return width, int(size[1] / size[0] * width)
    else:
        return int(size[0] / size[1] * height), height
